from amulet.api.history.base import RevisionManager
from ..data_types import EntryType


class RAMRevisionManager(RevisionManager):
    """A class to hold data about an entries history in RAM."""

    def _store_entry(self, entry: EntryType):
        self._revisions.append(entry)

    def get_current_entry(self):
        return self._revisions[self._current_revision_index]



from __future__ import annotations
from typing import Optional, Tuple, Generator, Set, Iterable, Dict
import weakref

from amulet.api.data_types import DimensionCoordinates, Dimension
from amulet.api.chunk import Chunk
from amulet.api.history.data_types import EntryType, EntryKeyType
from amulet.api.history.base import RevisionManager
from amulet.api.history.revision_manager import DBRevisionManager
from amulet.api.errors import ChunkDoesNotExist, ChunkLoadError
from amulet.api.history.history_manager import DatabaseHistoryManager
from amulet.api import level as api_level
from amulet.libs.leveldb import LevelDB


class ChunkDBEntry(DBRevisionManager):
    __slots__ = ("_world", "_history_db")

    def __init__(
        self,
        world: api_level.BaseLevel,
        history_db: LevelDB,
        prefix: str,
        initial_state: EntryType,
    ):
        self._world = weakref.ref(world)
        self._history_db = weakref.ref(history_db)
        super().__init__(prefix, initial_state)

    @property
    def world(self) -> api_level.BaseLevel:
        return self._world()

    def _serialise(self, path: str, entry: Optional[Chunk]) -> Optional[str]:
        if entry is None:
            return None
        else:
            pickled_bytes = entry.pickle()
            self._history_db().put(path.encode("utf-8"), pickled_bytes)
            return path

    def _deserialise(self, path: Optional[str]) -> Optional[Chunk]:
        if path is None:
            return None
        else:
            pickled_bytes = self._history_db().get(path.encode("utf-8"))
            return Chunk.unpickle(
                pickled_bytes, self.world.block_palette, self.world.biome_palette
            )


class ChunkManager(DatabaseHistoryManager):
    """
    The ChunkManager class is a class that handles chunks within a world.

    It handles the temporary database of chunks in RAM that can be directly modified.

    It handles a serialised version of the chunks on disk to reduce RAM usage.

    It also contains a history manager to allow undoing and redoing of changes.
    """

    _temporary_database: Dict[DimensionCoordinates, EntryType]
    _history_database: Dict[DimensionCoordinates, RevisionManager]

    DoesNotExistError = ChunkDoesNotExist
    LoadError = ChunkLoadError

    def __init__(self, level: api_level.BaseLevel, history_db: LevelDB):
        """
        Construct a :class:`ChunkManager` instance.

        Should not be directly used by third party code.

        :param level: The world that this chunk manager is associated with
        """
        super().__init__()
        self._prefix: str = f"chunks"  # the location to serialise Chunks to
        self._level = weakref.ref(level)
        self._history_db = weakref.ref(history_db)

    @property
    def level(self) -> api_level.BaseLevel:
        """The level that this chunk manager is associated with."""
        return self._level()

    def changed_chunks(self) -> Generator[DimensionCoordinates, None, None]:
        """The location of every chunk that has been changed since the last save."""
        yield from self.changed_entries()

    def unload(self, safe_area: Optional[Tuple[Dimension, int, int, int, int]] = None):
        """
        Unload all chunks from the temporary database that are not in the safe area.

        :param safe_area: The area that should not be unloaded [dimension, min_chunk_x, min_chunk_z, max_chunk_x, max_chunk_z]. If None will unload all chunk data.
        """
        with self._lock:
            if safe_area is None:
                self._temporary_database.clear()
            else:
                unload_chunks = []
                dimension, minx, minz, maxx, maxz = safe_area
                for (cd, cx, cz), chunk in self._temporary_database.items():
                    if not (
                        cd == dimension and minx <= cx <= maxx and minz <= cz <= maxz
                    ):
                        unload_chunks.append((cd, cx, cz))
                for chunk_key in unload_chunks:
                    del self._temporary_database[chunk_key]

    def __contains__(self, item: DimensionCoordinates) -> bool:
        """
        Is the chunk specified present in the level.

        >>> ("minecraft:overworld", 0, 0) in level.chunks
        True

        :param item: A tuple of dimension, chunk x coordinate and chunk z coordinate.
        :return: True if the chunk is present, False otherwise
        """
        return self._has_entry(item)

    def has_chunk(self, dimension: Dimension, cx: int, cz: int) -> bool:
        """
        Is the chunk specified present in the level.

        :param dimension: The dimension of the chunk to check.
        :param cx: The chunk x coordinate.
        :param cz: The chunk z coordinate.
        :return: True if the chunk is present, False otherwise
        """
        return self._has_entry((dimension, cx, cz))

    def _raw_has_entry(self, key: DimensionCoordinates):
        dimension, cx, cz = key
        return self.level.level_wrapper.has_chunk(cx, cz, dimension)

    def all_chunk_coords(self, dimension: Dimension) -> Set[Tuple[int, int]]:
        """
        The coordinates of every chunk in this world.

        This is the combination of chunks saved to the world and chunks yet to be saved.
        """
        return self._all_entries(dimension)

    def _all_entries(self, dimension: Dimension) -> Set[Tuple[int, int]]:
        # custom behaviour to handle the dimension option.
        with self._lock:
            coords = set()
            deleted_chunks = set()
            for key in self._temporary_database.keys():
                dim, cx, cz = key
                if dim == dimension:
                    if self._temporary_database[key] is None:
                        deleted_chunks.add((cx, cz))
                    else:
                        coords.add((cx, cz))

            for key in self._history_database.keys():
                dim, cx, cz = key
                if dim == dimension and key not in self._temporary_database:
                    if self._history_database[key].is_deleted:
                        deleted_chunks.add((cx, cz))
                    else:
                        coords.add((cx, cz))

            for key in self._raw_all_entries(dimension):
                if key not in coords and key not in deleted_chunks:
                    coords.add(key)

        return coords

    def _raw_all_entries(self, dimension: Dimension) -> Iterable[Tuple[int, int]]:
        return self.level.level_wrapper.all_chunk_coords(dimension)

    def get_chunk(self, dimension: Dimension, cx: int, cz: int) -> Chunk:
        """
        Gets the :class:`Chunk` object at the specified chunk coordinates.

        This may be a :class:`Chunk` instance if the chunk exists, None if it is known to not exist
        or :class:`~amulet.api.errors.ChunkDoesNotExist` will be raised if there is no record so it is unknown if it exists or not.

        Use has_chunk to check if there is a record of the chunk.

        :param dimension: The dimension to get the chunk from
        :param cx: The X coordinate of the desired chunk
        :param cz: The Z coordinate of the desired chunk
        :return: A Chunk instance or None
        :raises:
            :class:`~amulet.api.errors.ChunkDoesNotExist`: If the chunk does not exist (was deleted or never created)
        """
        return self._get_entry((dimension, cx, cz))

    def _raw_get_entry(self, key: EntryKeyType) -> EntryType:
        dimension, cx, cz = key
        chunk = self.level.level_wrapper.load_chunk(cx, cz, dimension)
        chunk.block_palette = self.level.block_palette
        chunk.biome_palette = self.level.biome_palette
        return chunk

    def put_chunk(self, chunk: Chunk, dimension: Dimension):
        """
        Add a given chunk to the chunk manager.

        :param chunk: The :class:`Chunk` to add to the chunk manager. It will be added at the location stored in :attr:`Chunk.coordinates`
        :param dimension: The dimension to add the chunk to.
        """
        chunk.block_palette = self.level.block_palette
        chunk.biome_palette = self.level.biome_palette
        self._put_entry((dimension, chunk.cx, chunk.cz), chunk)

    def delete_chunk(self, dimension: Dimension, cx: int, cz: int):
        """
        Delete a chunk from the chunk manager.

        :param cx: The X coordinate of the chunk
        :param cz: The Z coordinate of the chunk
        :param dimension: The dimension to delete the chunk from.
        """
        self._delete_entry((dimension, cx, cz))

    def _create_new_revision_manager(
        self, key: EntryKeyType, original_entry: EntryType
    ) -> RevisionManager:
        dimension, cx, cz = key
        prefix = f"{self._prefix}/{dimension}/{cx}.{cz}"
        return ChunkDBEntry(self.level, self._history_db(), prefix, original_entry)



from __future__ import annotations
from typing import Tuple, Dict, Optional
import numpy

from amulet_nbt import (
    CompoundTag,
    ListTag,
    ByteTag,
    IntTag,
    LongTag,
    LongArrayTag,
    StringTag,
)

from amulet.api.chunk import Chunk
from amulet.api.registry import BiomeManager
from amulet.api.data_types import AnyNDArray, BiomeType
from amulet.utils.world_utils import (
    decode_long_array,
    encode_long_array,
)

from .base_anvil_interface import (
    ChunkDataType,
    ChunkPathType,
)
from .anvil_2709 import (
    Anvil2709Interface as ParentInterface,
)


class Anvil2844Interface(ParentInterface):
    """
    Note that some of these changes happened in earlier snapshots
    Chunk restructuring
    Contents of Level tag moved into root
    Some tags renamed from PascalCase to snake_case
        2844
        Level.Entities -> entities.
        Level.TileEntities -> block_entities.
        Level.TileTicks and Level.ToBeTicked have moved to block_ticks.
        Level.LiquidTicks and Level.LiquidsToBeTicked have moved to fluid_ticks.
        Level.Sections -> sections.
        Level.Structures -> structures.
        Level.Structures.Starts -> structures.starts.
        Level.Sections[].block_states -> sections[].block_states.
        Level.Sections[].biomes -> sections[].biomes
        Added yPos the minimum section y position in the chunk.
        Added below_zero_retrogen containing data to support below zero generation.
        Added blending_data containing data to support blending new world generation with existing chunks.
        2836
        Level.Sections[].BlockStates & Level.Sections[].Palette -> Level.Sections[].block_states.
        Level.Biomes -> Level.Sections[].biomes.
        Level.CarvingMasks[] is now long[] instead of byte[].
    """

    Level: ChunkPathType = ("region", [], CompoundTag)
    Sections: ChunkPathType = ("region", [("sections", ListTag)], ListTag)

    BlockEntities: ChunkPathType = ("region", [("block_entities", ListTag)], ListTag)
    Entities: ChunkPathType = ("region", [("entities", ListTag)], ListTag)
    InhabitedTime: ChunkPathType = ("region", [("InhabitedTime", LongTag)], LongTag)
    LastUpdate: ChunkPathType = ("region", [("LastUpdate", LongTag)], LongTag)
    Heightmaps: ChunkPathType = ("region", [("Heightmaps", CompoundTag)], CompoundTag)
    BlockTicks: ChunkPathType = ("region", [("block_ticks", ListTag)], ListTag)
    ToBeTicked = None
    Biomes = None

    Status: ChunkPathType = ("region", [("Status", StringTag)], StringTag("full"))

    LiquidTicks: ChunkPathType = ("region", [("fluid_ticks", ListTag)], ListTag)
    PostProcessing: ChunkPathType = ("region", [("PostProcessing", ListTag)], ListTag)

    Structures: ChunkPathType = ("region", [("structures", CompoundTag)], CompoundTag)

    yPos: ChunkPathType = ("region", [("yPos", IntTag)], IntTag)

    @staticmethod
    def minor_is_valid(key: int):
        return 2844 <= key < 3150

    def _get_floor_cy(self, data: ChunkDataType):
        return self.get_layer_obj(data, self.yPos, pop_last=True).py_int

    def _decode_block_section(
        self, section: CompoundTag
    ) -> Optional[Tuple[numpy.ndarray, list]]:
        block_states = self.get_obj(section, "block_states", CompoundTag)
        if (
            isinstance(block_states, CompoundTag) and "palette" in block_states
        ):  # 1.14 makes block_palette/blocks optional.
            section_palette = self._decode_block_palette(block_states.pop("palette"))
            data = block_states.pop("data", None)
            if data is None:
                arr = numpy.zeros((16, 16, 16), numpy.uint32)
            else:
                decoded = decode_long_array(
                    data.np_array,
                    16**3,
                    max(4, (len(section_palette) - 1).bit_length()),
                    dense=self.LongArrayDense,
                ).astype(numpy.uint32)
                arr = numpy.transpose(decoded.reshape((16, 16, 16)), (2, 0, 1))
            return arr, section_palette
        else:
            return None

    @staticmethod
    def _decode_biome_palette(palette: ListTag) -> list[BiomeType]:
        return [entry.py_data for entry in palette]

    def _decode_biome_section(
        self, section: CompoundTag
    ) -> Optional[Tuple[numpy.ndarray, list]]:
        biomes = self.get_obj(section, "biomes", CompoundTag)
        if isinstance(biomes, CompoundTag) and "palette" in biomes:
            section_palette = self._decode_biome_palette(biomes.pop("palette"))
            assert section_palette, "Biome palette cannot be empty"
            data = biomes.pop("data", None)
            if data is None:
                # case 1: palette contains one value and data does not exist (undefined zero array)
                # TODO: in the new biome system just leave this as the number
                arr = numpy.zeros((4, 4, 4), numpy.uint32)
            else:
                # case 2: palette contains values and data is an index array
                arr = numpy.transpose(
                    decode_long_array(
                        data.np_array,
                        4**3,
                        max(1, (len(section_palette) - 1).bit_length()),
                        dense=self.LongArrayDense,
                    )
                    .astype(numpy.uint32)
                    .reshape((4, 4, 4)),
                    (2, 0, 1),
                )
            return arr, section_palette
        else:
            return None

    def _decode_biomes(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        biomes: Dict[int, numpy.ndarray] = {}
        palette = BiomeManager()

        for cy, section in self._iter_sections(data):
            data = self._decode_biome_section(section)
            if data is not None:
                arr, section_palette = data
                lut = numpy.array(
                    [palette.get_add_biome(biome) for biome in section_palette]
                )
                biomes[cy] = lut[arr].astype(numpy.uint32)

        chunk.biomes = biomes
        chunk.biome_palette = palette

    def _decode_block_ticks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc.setdefault("block_ticks", {}).update(
            self._decode_ticks(self.get_layer_obj(data, self.BlockTicks, pop_last=True))
        )

    def _decode_fluid_ticks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc.setdefault("fluid_ticks", {}).update(
            self._decode_ticks(
                self.get_layer_obj(data, self.LiquidTicks, pop_last=True)
            )
        )

    def _encode_coords(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        super()._encode_coords(chunk, data, floor_cy, height_cy)
        self.set_layer_obj(data, self.yPos, IntTag(floor_cy))

    def _encode_block_section(
        self,
        chunk: Chunk,
        sections: Dict[int, CompoundTag],
        palette: AnyNDArray,
        cy: int,
    ):
        block_sub_array = numpy.transpose(
            chunk.blocks.get_sub_chunk(cy), (1, 2, 0)
        ).ravel()

        sub_palette_, block_sub_array = numpy.unique(
            block_sub_array, return_inverse=True
        )
        sub_palette = self._encode_block_palette(palette[sub_palette_])
        section = sections.setdefault(cy, CompoundTag())
        block_states = section["block_states"] = CompoundTag({"palette": sub_palette})
        if len(sub_palette) != 1:
            block_states["data"] = LongArrayTag(
                encode_long_array(
                    block_sub_array, dense=self.LongArrayDense, min_bits_per_entry=4
                )
            )

    @staticmethod
    def _encode_biome_palette(palette: list[BiomeType]) -> ListTag:
        return ListTag([StringTag(entry) for entry in palette])

    def _encode_biome_section(
        self,
        chunk: Chunk,
        sections: Dict[int, CompoundTag],
        cy: int,
    ):
        biome_sub_array = numpy.transpose(
            chunk.biomes.get_section(cy), (1, 2, 0)
        ).ravel()

        sub_palette_, biome_sub_array = numpy.unique(
            biome_sub_array, return_inverse=True
        )
        sub_palette = self._encode_biome_palette(chunk.biome_palette[sub_palette_])
        biomes = sections[cy]["biomes"] = CompoundTag({"palette": sub_palette})
        if len(sub_palette) != 1:
            biomes["data"] = LongArrayTag(
                encode_long_array(biome_sub_array, dense=self.LongArrayDense)
            )

    def _encode_biomes(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        sections = self._get_encode_sections(data, floor_cy, height_cy)
        ceil_cy = floor_cy + height_cy
        chunk.biomes.convert_to_3d()
        for cy in chunk.biomes.sections:
            if floor_cy <= cy < ceil_cy:
                self._encode_biome_section(chunk, sections, cy)

    def _encode_block_ticks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data, self.BlockTicks, self._encode_ticks(chunk.misc.get("block_ticks", {}))
        )

    def _encode_fluid_ticks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data,
            self.LiquidTicks,
            self._encode_ticks(chunk.misc.get("fluid_ticks", {})),
        )


export = Anvil2844Interface



from __future__ import annotations

import numpy
from typing import TYPE_CHECKING

from amulet.level.translators.chunk.bedrock import BaseBedrockTranslator
from amulet.api.data_types import BlockNDArray, AnyNDArray

if TYPE_CHECKING:
    from amulet.api.block import Block
    from PyMCTranslate import Version


class BedrockPsudoNumericalTranslator(BaseBedrockTranslator):
    @staticmethod
    def is_valid(key):
        return key[0] == "bedrock" and (1, 2, 13) <= key[1] < (1, 13, 0)

    def _pack_block_palette(
        self, version: "Version", palette: BlockNDArray
    ) -> AnyNDArray:
        """
        Packs a numpy array of Block objects into an object array containing one pair of version number and Block objects.
        :param version:
        :param palette:
        :return: numpy.ndarray[Tuple[Tuple[Optional[VersionNumber], Block]]]
        """
        palette_ = numpy.empty(len(palette), dtype=object)
        for palette_index, block in enumerate(palette):
            block: "Block"
            # TODO: perhaps check that property 'block_data' exists and is IntTag user interaction if not
            palette_[palette_index] = tuple((None, b) for b in block.block_tuple)

        return palette_


export = BedrockPsudoNumericalTranslator



from __future__ import annotations

from typing import Tuple, Dict, List, Union, Iterable, Optional, TYPE_CHECKING, Any
import struct
import logging

import numpy
from amulet_nbt import (
    ShortTag,
    IntTag,
    StringTag,
    CompoundTag,
    NamedTag,
    load as load_nbt,
    load_many,
    ReadContext,
    utf8_escape_decoder,
    utf8_escape_encoder,
)

import amulet
from amulet.api.block import Block
from amulet.api.chunk import Chunk, StatusFormats

from amulet.utils.numpy_helpers import brute_sort_objects
from amulet.utils.world_utils import fast_unique, from_nibble_array
from amulet.api.wrapper import Interface
from amulet.api.data_types import (
    AnyNDArray,
    SubChunkNDArray,
    PlatformType,
    VersionNumberTuple,
    VersionIdentifierTuple,
)
from amulet.level import loader
from amulet.api.wrapper import EntityIDType, EntityCoordType
from .leveldb_chunk_versions import chunk_to_game_version
from amulet.api.chunk.entity_list import EntityList
from amulet.level.formats.leveldb_world.chunk import ChunkData

if TYPE_CHECKING:
    from amulet.api.block_entity import BlockEntity
    from amulet.api.entity import Entity
    from amulet.api.chunk.blocks import Blocks
    from amulet.api.wrapper import Translator

log = logging.getLogger(__name__)

# This is here to scale a 4x array to a 16x array. This can be removed when we natively support 16x array
_scale_grid = tuple(numpy.meshgrid(*[numpy.arange(16) // 4] * 3, indexing="ij"))


class BaseLevelDBInterface(Interface):

    chunk_version: int = None

    def __init__(self):
        self._feature_options = {
            "finalised_state": ["int0-2"],
            "data_2d": ["height512|biome256", "height512|biome4096"],
            "block_entities": ["31list"],
            "block_entity_format": [EntityIDType.namespace_str_id, EntityIDType.str_id],
            "block_entity_coord_format": [EntityCoordType.xyz_int],
            "entities": ["32list", "actor"],
            "entity_format": [
                EntityIDType.namespace_str_identifier,
                EntityIDType.int_id,
            ],
            "entity_coord_format": [EntityCoordType.Pos_list_float],
            "terrain": ["30array", "2farray", "2f1palette", "2fnpalette"],
        }
        self._features = {key: None for key in self._feature_options.keys()}

    def _set_feature(self, feature: str, option: Any):
        assert feature in self._feature_options, f"{feature} is not a valid feature."
        assert (
            option in self._feature_options[feature]
        ), f'Invalid option {option} for feature "{feature}"'
        self._features[feature] = option

    def is_valid(self, key: Tuple) -> bool:
        return key[0] == "bedrock" and key[1] == self.chunk_version

    def get_translator(
        self,
        max_world_version: Tuple[PlatformType, VersionNumberTuple],
        data: ChunkData = None,
    ) -> Tuple["Translator", VersionNumberTuple]:
        """
        Get the Translator class for the requested version.
        :param max_world_version: The game version the world was last opened in. Version number tuple or data version number.
        :param data: Optional data to get translator based on chunk version rather than world version
        :return: Tuple[Translator, version number for PyMCTranslate to use]
        """
        if data:
            if b"," in data:
                chunk_version = data[b","][0]
            else:
                chunk_version = data[b"v"][0]
            game_version = chunk_to_game_version(max_world_version[1], chunk_version)
        else:
            game_version = max_world_version[1]
        return loader.Translators.get(("bedrock", game_version)), game_version

    @staticmethod
    def _chunk_key_to_sub_chunk(cy: int, min_y: int) -> int:
        """Convert the database sub-chunk key to the sub-chunk index."""
        return cy

    @staticmethod
    def _get_sub_chunk_storage_byte(cy: int, min_y: int) -> bytes:
        return struct.pack("b", cy)

    def decode(
        self, cx: int, cz: int, chunk_data: ChunkData, bounds: Tuple[int, int]
    ) -> Tuple[Chunk, AnyNDArray]:
        """
        Create an amulet.api.chunk.Chunk object from raw data given by the format
        :param cx: chunk x coordinate
        :param cz: chunk z coordinate
        :param chunk_data: Raw chunk data provided by the format.
        :param bounds: The minimum and maximum height of the chunk.
        :return: Chunk object in version-specific format, along with the block_palette for that chunk.
        """
        chunk = Chunk(cx, cz)
        chunk_palette = numpy.empty(0, dtype=object)
        chunk.misc = {"bedrock_chunk_data": chunk_data}

        chunk_data.pop(b"v", None)
        chunk_data.pop(b",", None)

        if self._features["terrain"].startswith(
            "2f"
        ):  # ["2farray", "2f1palette", "2fnpalette"]
            subchunks = {}
            for key in chunk_data.copy().keys():
                if len(key) == 2 and key[0:1] == b"\x2F":
                    cy = struct.unpack("b", key[1:2])[0]
                    subchunks[
                        self._chunk_key_to_sub_chunk(cy, bounds[0] >> 4)
                    ] = chunk_data.pop(key)
            chunk.blocks, chunk_palette = self._load_subchunks(subchunks)
        elif self._features["terrain"] == "30array":
            section_data = chunk_data.pop(b"\x30", None)
            if section_data is not None:
                block_ids = numpy.frombuffer(
                    section_data[: 2**15], dtype=numpy.uint8
                ).astype(numpy.uint16)
                block_data = from_nibble_array(
                    numpy.frombuffer(
                        section_data[2**15 : 2**15 + 2**14], dtype=numpy.uint8
                    )
                )

                # there is other data here but we are going to skip over it
                combined_palette, block_array = fast_unique(
                    numpy.transpose(
                        ((block_ids << 4) + block_data).reshape(16, 16, 128), (0, 2, 1)
                    )
                )
                chunk.blocks = {
                    i: block_array[:, i * 16 : (i + 1) * 16, :] for i in range(8)
                }
                palette: AnyNDArray = numpy.array(
                    [combined_palette >> 4, combined_palette & 15]
                ).T
                chunk_palette = numpy.empty(len(palette), dtype=object)
                for i, b in enumerate(palette):
                    chunk_palette[i] = ((None, tuple(b)),)

        else:
            raise Exception

        if self._features["finalised_state"] == "int0-2":
            state = chunk_data.pop(b"\x36", None)
            val = 2
            if isinstance(state, bytes):
                if len(state) == 1:
                    # old versions of the game store this as a byte
                    val = struct.unpack("b", state)[0]
                elif len(state) == 4:
                    # newer versions store it as an int
                    val = struct.unpack("<i", state)[0]
            chunk.status = val

        if b"+" in chunk_data:
            height, biome = self._decode_height_3d_biomes(
                chunk_data[b"+"], bounds[0] >> 4
            )
            chunk.misc["height"] = height
            chunk.biomes = biome
        elif b"\x2D" in chunk_data:
            d2d = chunk_data[b"\x2D"]
            height, biome = (
                numpy.frombuffer(d2d[:512], "<i2").reshape((16, 16)),
                d2d[512:],
            )
            chunk.misc["height"] = height
            chunk.biomes = numpy.frombuffer(biome, dtype="uint8").reshape(16, 16).T

        # TODO: implement key support
        # \x2D  heightmap and biomes
        # \x31  block entity
        # \x32  entity
        # \x33  ticks
        # \x34  block extra data
        # \x35  biome state
        # \x39  7 ints and an end (03)? Honestly don't know what this is
        # \x3A  fire tick?

        # \x2E  2d legacy
        # \x30  legacy terrain

        # unpack block entities and entities
        if self._features["block_entities"] == "31list":
            block_entities = self._unpack_nbt_list(chunk_data.pop(b"\x31", b""))
            chunk.block_entities = self._decode_block_entity_list(block_entities)

        if self._features["entities"] in ("32list", "actor"):
            entities = self._unpack_nbt_list(chunk_data.pop(b"\x32", b""))
            if amulet.entity_support:
                chunk.entities = self._decode_entity_list(entities)
            else:
                chunk._native_entities = EntityList(self._decode_entity_list(entities))
                chunk._native_version = (
                    "bedrock",
                    (0, 0, 0),
                )  # TODO: find a way to determine entity version

        if self._features["entities"] == "actor":
            if chunk_data.entity_actor:
                if amulet.entity_support:
                    chunk.entities += self._decode_entity_list(chunk_data.entity_actor)
                else:
                    chunk._native_entities += self._decode_entity_list(
                        chunk_data.entity_actor
                    )
                    chunk._native_version = ("bedrock", (0, 0, 0))
                chunk_data.entity_actor.clear()

        return chunk, chunk_palette

    def encode(
        self,
        chunk: Chunk,
        palette: AnyNDArray,
        max_world_version: VersionIdentifierTuple,
        bounds: Tuple[int, int],
    ) -> Dict[bytes, Optional[bytes]]:
        chunk_data = chunk.misc.get("bedrock_chunk_data", {})
        if isinstance(chunk_data, ChunkData):
            pass
        elif isinstance(chunk_data, dict):
            chunk_data = ChunkData(
                {
                    k: v
                    for k, v in chunk_data.items()
                    if isinstance(k, bytes) and isinstance(v, bytes)
                }
            )
        else:
            chunk_data = ChunkData()

        chunk_data: ChunkData

        # chunk version
        if self.chunk_version is not None:
            chunk_data[b"v" if self.chunk_version <= 20 else b","] = bytes(
                [self.chunk_version]
            )

        # terrain data
        terrain = self._encode_subchunks(
            chunk.blocks, palette, bounds, max_world_version
        )
        min_y = bounds[0] // 16
        for cy, sub_chunk in terrain.items():
            chunk_data[
                b"\x2F" + self._get_sub_chunk_storage_byte(cy, min_y)
            ] = sub_chunk

        # chunk status
        if self._features["finalised_state"] == "int0-2":
            chunk_data[b"\x36"] = struct.pack(
                "<i", chunk.status.as_type(StatusFormats.Bedrock)
            )

        # biome and height data
        if self._features["data_2d"] == "height512|biome256":
            d2d: List[bytes] = [self._encode_height(chunk)]
            chunk.biomes.convert_to_2d()
            d2d.append(chunk.biomes.astype("uint8").T.tobytes())
            chunk_data[b"\x2D"] = b"".join(d2d)
            if b"+" in chunk_data:
                chunk_data[b"+"] = None
        elif self._features["data_2d"] == "height512|biome4096":
            chunk_data[b"+"] = self._encode_height_3d_biomes(
                chunk, bounds[0] >> 4, bounds[1] >> 4
            )
            if b"\x2D" in chunk_data:
                chunk_data[b"\x2D"] = None

        # pack block entities and entities
        if self._features["block_entities"] == "31list":
            block_entities_out = self._encode_block_entity_list(chunk.block_entities)

            if block_entities_out:
                chunk_data[b"\x31"] = self._pack_nbt_list(block_entities_out)
            else:
                chunk_data[b"\x31"] = None

        if self._features["entities"] == "32list":

            def save_entities(entities_out):
                if entities_out:
                    chunk_data[b"\x32"] = self._pack_nbt_list(entities_out)
                else:
                    chunk_data[b"\x32"] = None

            if amulet.entity_support:
                save_entities(self._encode_entity_list(chunk.entities))
            else:
                try:
                    if chunk._native_version[0] == "bedrock":
                        save_entities(self._encode_entity_list(chunk._native_entities))
                except:
                    pass

        elif self._features["entities"] == "actor":
            chunk_data.entity_actor.clear()
            if amulet.entity_support:
                chunk_data.entity_actor.extend(self._encode_entity_list(chunk.entities))
            else:
                try:
                    if chunk._native_version[0] == "bedrock":
                        chunk_data.entity_actor.extend(
                            self._encode_entity_list(chunk._native_entities)
                        )
                except:
                    pass

        return chunk_data

    def _load_subchunks(
        self, subchunks: Dict[int, Optional[bytes]]
    ) -> Tuple[Dict[int, SubChunkNDArray], AnyNDArray]:
        """
        Load a list of bytes objects which contain chunk data
        This function should be able to load all sub-chunk formats (technically before it)
        All sub-chunks will almost certainly all have the same sub-chunk version but
        it should be able to handle a case where that is not true.

        As such this function will return a Chunk and a rather complicated block_palette
        The newer formats allow multiple blocks to occupy the same space and the
        newer versions also include a version ber block. So this will also need
        returning for the translator to handle.

        The block_palette will be a numpy array containing tuple objects
            The tuple represents the "block" however can contain more than one Block object.
            Inside the tuple are one or more tuples.
                These include the block version number and the block itself
                    The block version number will be either None if no block version is given
                    or a tuple containing 4 ints.

                    The block will be either a Block class for the newer formats or a tuple of two ints for the older formats
        """
        blocks: Dict[int, SubChunkNDArray] = {}
        palette: List[
            Tuple[Tuple[Optional[int], Union[Tuple[int, int], Block]], ...]
        ] = [
            (
                (
                    17563649,
                    Block(
                        namespace="minecraft",
                        base_name="air",
                        properties={"block_data": IntTag(0)},
                    ),
                ),
            )
        ]
        for cy, data in subchunks.items():
            if data is None:
                continue
            if data[0] in {0, 2, 3, 4, 5, 6, 7}:
                block_ids = numpy.frombuffer(
                    data[1 : 1 + 2**12], dtype=numpy.uint8
                ).astype(numpy.uint16)
                block_data = from_nibble_array(
                    numpy.frombuffer(
                        data[1 + 2**12 : 1 + 2**12 + 2**11], dtype=numpy.uint8
                    )
                )
                combined_palette, block_array = fast_unique(
                    numpy.transpose(
                        ((block_ids << 4) + block_data).reshape(16, 16, 16), (0, 2, 1)
                    )
                )
                blocks[cy] = block_array + len(palette)
                for b in numpy.array([combined_palette >> 4, combined_palette & 15]).T:
                    palette.append(((None, tuple(b)),))

            else:
                if data[0] == 1:
                    storage_count = 1
                    data = data[1:]
                elif data[0] == 8:
                    storage_count, data = data[1], data[2:]
                elif data[0] == 9:
                    # There is an extra byte in this format storing the cy value
                    storage_count, cy, data = (
                        data[1],
                        struct.unpack("b", data[2:3])[0],
                        data[3:],
                    )
                else:
                    raise Exception(f"sub-chunk version {data[0]} is not known.")

                sub_chunk_blocks = numpy.zeros(
                    (16, 16, 16, storage_count), dtype=numpy.uint32
                )
                sub_chunk_palette: List[List[Tuple[Optional[int], Block]]] = []
                for storage_index in range(storage_count):
                    (
                        sub_chunk_blocks[:, :, :, storage_index],
                        palette_data,
                        data,
                    ) = self._load_palette_blocks(data)
                    palette_data_out: List[Tuple[Optional[int], Block]] = []
                    for block in palette_data:
                        block = block.compound
                        namespace, base_name = block["name"].py_str.split(":", 1)
                        if "version" in block:
                            version: Optional[int] = block.get_int("version").py_int
                        else:
                            version = None

                        if "states" in block or "val" not in block:  # 1.13 format
                            properties = block.get_compound(
                                "states", CompoundTag()
                            ).py_dict
                            if version is None:
                                version = 17694720  # 1, 14, 0, 0
                        else:
                            properties = {"block_data": IntTag(block["val"].py_int)}
                        palette_data_out.append(
                            (
                                version,
                                Block(
                                    namespace=namespace,
                                    base_name=base_name,
                                    properties=properties,
                                ),
                            )
                        )
                    sub_chunk_palette.append(palette_data_out)

                if storage_count == 1:
                    blocks[cy] = sub_chunk_blocks[:, :, :, 0] + len(palette)
                    palette += [(val,) for val in sub_chunk_palette[0]]
                elif storage_count > 1:
                    # we have two or more storages so need to find the unique block combinations and merge them together
                    sub_chunk_palette_, sub_chunk_blocks = numpy.unique(
                        sub_chunk_blocks.reshape(-1, storage_count),
                        return_inverse=True,
                        axis=0,
                    )
                    blocks[cy] = sub_chunk_blocks.reshape(16, 16, 16).astype(
                        numpy.uint32
                    ) + len(palette)
                    palette += [
                        tuple(
                            sub_chunk_palette[storage_index][index]
                            for storage_index, index in enumerate(palette_indexes)
                            if not (
                                storage_index > 0
                                and sub_chunk_palette[storage_index][index][
                                    1
                                ].namespaced_name
                                == "minecraft:air"
                            )
                        )
                        for palette_indexes in sub_chunk_palette_
                    ]
                else:
                    continue

        # block_palette should now look like this
        # List[
        #   Tuple[
        #       Tuple[version, Block], ...
        #   ]
        # ]

        numpy_palette, lut = brute_sort_objects(palette)
        for cy in blocks.keys():
            blocks[cy] = lut[blocks[cy]]

        return blocks, numpy_palette

    def _encode_subchunks(
        self,
        blocks: "Blocks",
        palette: AnyNDArray,
        bounds: Tuple[int, int],
        max_world_version: VersionIdentifierTuple,
    ) -> Dict[int, Optional[bytes]]:
        raise NotImplementedError

    def _save_subchunks_1(
        self, blocks: "Blocks", palette: AnyNDArray
    ) -> Dict[int, Optional[bytes]]:
        for index, block in enumerate(palette):
            block: Tuple[Tuple[None, Block], ...]
            block_data_tag = block[0][1].properties.get("block_data", IntTag(0))
            if isinstance(block_data_tag, IntTag):
                block_data = block_data_tag.py_int
                # if block_data >= 16:
                #     block_data = 0
            else:
                block_data = 0

            palette[index] = NamedTag(
                CompoundTag(
                    {
                        "name": StringTag(block[0][1].namespaced_name),
                        "val": ShortTag(block_data),
                    }
                )
            )
        chunk = {}
        for cy in range(16):
            if cy in blocks:
                palette_index, sub_chunk = fast_unique(blocks.get_sub_chunk(cy))
                sub_chunk_palette = list(palette[palette_index])
                chunk[cy] = b"\x01" + self._save_palette_subchunk(
                    sub_chunk.ravel(), sub_chunk_palette
                )
            else:
                chunk[cy] = None
        return chunk

    # These arent actual blocks, just ids pointing to the block_palette.

    @staticmethod
    def _decode_packed_array(data: bytes) -> Tuple[bytes, int, Optional[numpy.ndarray]]:
        """
        Parse a packed array as documented here
        https://gist.github.com/Tomcc/a96af509e275b1af483b25c543cfbf37

        :param data: The data to parse
        :return:
        """
        # Ignore LSB of data (its a flag) and get compacting level
        bits_per_value, data = struct.unpack("b", data[0:1])[0] >> 1, data[1:]
        if bits_per_value > 0:
            values_per_word = (
                32 // bits_per_value
            )  # Word = 4 bytes, basis of compacting.
            word_count = -(
                -4096 // values_per_word
            )  # Ceiling divide is inverted floor divide

            arr = numpy.packbits(
                numpy.pad(
                    numpy.unpackbits(
                        numpy.frombuffer(
                            bytes(reversed(data[: 4 * word_count])), dtype="uint8"
                        )
                    )
                    .reshape(-1, 32)[:, -values_per_word * bits_per_value :]
                    .reshape(-1, bits_per_value)[-4096:, :],
                    [(0, 0), (16 - bits_per_value, 0)],
                    "constant",
                )
            ).view(dtype=">i2")[::-1]
            arr = arr.reshape((16, 16, 16)).swapaxes(1, 2)
            data = data[4 * word_count :]
        else:
            arr = None
        return data, bits_per_value, arr

    def _decode_height_3d_biomes(
        self, data: bytes, floor_cy: int
    ) -> Tuple[numpy.ndarray, Dict[int, numpy.ndarray]]:
        # The 3D biome format consists of 25 16x arrays with the first array corresponding to the lowest sub-chunk in the world
        #  This is -64 in the overworld and 0 in the nether and end
        # TODO: make this support the full 16x
        heightmap, data = (
            numpy.frombuffer(data[:512], "<i2").reshape((16, 16)),
            data[512:],
        )
        biomes = {}
        cy = floor_cy
        while data:
            data, bits_per_value, arr = self._decode_packed_array(data)
            if bits_per_value == 0:
                value, data = struct.unpack(f"<I", data[:4])[0], data[4:]
                # TODO: when the new biome system supports ints just return the value
                biomes[cy] = numpy.full((4, 4, 4), value, dtype=numpy.uint32)
            elif bits_per_value > 0:
                arr = arr[::4, ::4, ::4]
                palette_len, data = struct.unpack("<I", data[:4])[0], data[4:]
                biomes[cy] = numpy.frombuffer(data, "<i4", palette_len)[arr].astype(
                    numpy.uint32
                )
                data = data[4 * palette_len :]
            cy += 1
        return heightmap, biomes

    def _encode_height(self, chunk) -> bytes:
        height = chunk.misc.get("height")
        if isinstance(height, numpy.ndarray) and height.size == 256:
            return height.ravel().astype("<i2").tobytes()
        else:
            return b"\x00" * 512

    def _encode_height_3d_biomes(
        self, chunk: Chunk, floor_cy: int, ceil_cy: int
    ) -> bytes:
        d2d: List[bytes] = [self._encode_height(chunk)]
        chunk.biomes.convert_to_3d()
        # at least one biome array needs to be defined
        # all biome arrays below the highest biome array must be populated.
        highest = next(
            (cy for cy in range(ceil_cy, floor_cy - 1, -1) if cy in chunk.biomes), None
        )
        if highest is None:
            # populate lowest array
            chunk.biomes.create_section(floor_cy)
        else:
            for cy in range(highest - 1, floor_cy - 1, -1):
                if cy not in chunk.biomes:
                    chunk.biomes.add_section(
                        cy,
                        numpy.repeat(
                            # get the array for the sub-chunk above and get the lowest slice
                            chunk.biomes.get_section(cy + 1)[:, :1, :],
                            4,  # Repeat this slice 4 times in the first axis
                            1,  # TODO: When biome editing supports 16x this will need to be changed.
                        ),
                    )

        for cy in range(floor_cy, floor_cy + 25):
            if cy in chunk.biomes:
                arr = chunk.biomes.get_section(cy)
                palette, arr_uniq = numpy.unique(arr, return_inverse=True)
                if len(palette) == 1:
                    d2d.append(b"\x01")
                else:
                    d2d.append(
                        self._encode_packed_array(
                            arr_uniq.reshape(arr.shape)[_scale_grid]
                        )
                    )
                    d2d.append(struct.pack("<I", len(palette)))
                d2d.append(palette.astype("<i4").tobytes())
            else:
                d2d.append(b"\xFF")

        return b"".join(d2d)

    def _load_palette_blocks(
        self,
        data: bytes,
    ) -> Tuple[numpy.ndarray, List[NamedTag], bytes]:
        data, _, blocks = self._decode_packed_array(data)
        if blocks is None:
            blocks = numpy.zeros((16, 16, 16), dtype=numpy.int16)
            palette_len = 1
        else:
            palette_len, data = struct.unpack("<I", data[:4])[0], data[4:]

        if palette_len:
            read_context = ReadContext()
            palette = load_many(
                data,
                compressed=False,
                count=palette_len,
                little_endian=True,
                read_context=read_context,
            )
            data = data[read_context.offset :]
        else:
            palette = [
                NamedTag(
                    CompoundTag(
                        {
                            "name": StringTag("minecraft:air"),
                            "states": CompoundTag(),
                            "version": IntTag(17694723),
                        }
                    )
                )
            ]

        return blocks, palette, data

    @staticmethod
    def _encode_packed_array(arr: numpy.ndarray, min_bit_size=1) -> bytes:
        bits_per_value = max(int(numpy.amax(arr)).bit_length(), min_bit_size)
        if bits_per_value == 7:
            bits_per_value = 8
        elif 9 <= bits_per_value <= 15:
            bits_per_value = 16
        header = bytes([bits_per_value << 1])

        values_per_word = 32 // bits_per_value  # Word = 4 bytes, basis of compacting.
        word_count = -(
            -4096 // values_per_word
        )  # Ceiling divide is inverted floor divide

        arr = arr.swapaxes(1, 2).ravel()
        packed_arr = bytes(
            reversed(
                numpy.packbits(
                    numpy.pad(
                        numpy.pad(
                            numpy.unpackbits(
                                numpy.ascontiguousarray(arr[::-1], dtype=">i").view(
                                    dtype="uint8"
                                )
                            ).reshape(4096, -1)[:, -bits_per_value:],
                            [(word_count * values_per_word - 4096, 0), (0, 0)],
                            "constant",
                        ).reshape(-1, values_per_word * bits_per_value),
                        [(0, 0), (32 - values_per_word * bits_per_value, 0)],
                        "constant",
                    )
                )
                .view(dtype=">i4")
                .tobytes()
            )
        )
        return header + packed_arr

    def _save_palette_subchunk(
        self, blocks: numpy.ndarray, palette: List[NamedTag]
    ) -> bytes:
        """Save a single layer of blocks in the block_palette format"""
        return b"".join(
            [self._encode_packed_array(blocks), struct.pack("<I", len(palette))]
            + [
                block.save_to(
                    compressed=False,
                    little_endian=True,
                    string_encoder=utf8_escape_encoder,
                )
                for block in palette
            ]
        )

    @staticmethod
    def _unpack_nbt_list(raw_nbt: bytes) -> List[NamedTag]:
        nbt_list = []
        while raw_nbt:
            read_context = ReadContext()
            nbt = load_nbt(
                raw_nbt,
                little_endian=True,
                read_context=read_context,
                string_decoder=utf8_escape_decoder,
            )
            raw_nbt = raw_nbt[read_context.offset :]
            nbt_list.append(nbt)
        return nbt_list

    @staticmethod
    def _pack_nbt_list(nbt_list: List[NamedTag]):
        return b"".join(
            [
                nbt.save_to(
                    compressed=False,
                    little_endian=True,
                    string_encoder=utf8_escape_encoder,
                )
                for nbt in nbt_list
                if isinstance(nbt, NamedTag)
            ]
        )

    def _decode_entity_list(self, entities: List[NamedTag]) -> List["Entity"]:
        entities_out = []
        for nbt in entities:
            entity = self._decode_entity(
                nbt,
                self._features["entity_format"],
                self._features["entity_coord_format"],
            )
            if entity is not None:
                entities_out.append(entity)

        return entities_out

    def _encode_entity_list(self, entities: "EntityList") -> List[NamedTag]:
        entities_out = []
        for entity in entities:
            nbt = self._encode_entity(
                entity,
                self._features["entity_format"],
                self._features["entity_coord_format"],
            )
            if nbt is not None:
                entities_out.append(nbt)

        return entities_out

    def _decode_block_entity_list(
        self, block_entities: List[NamedTag]
    ) -> List["BlockEntity"]:
        entities_out = []
        for nbt in block_entities:
            entity = self._decode_block_entity(
                nbt,
                self._features["block_entity_format"],
                self._features["block_entity_coord_format"],
            )
            if entity is not None:
                entities_out.append(entity)

        return entities_out

    def _encode_block_entity_list(
        self, block_entities: Iterable["BlockEntity"]
    ) -> List[NamedTag]:
        entities_out = []
        for entity in block_entities:
            nbt = self._encode_block_entity(
                entity,
                self._features["block_entity_format"],
                self._features["block_entity_coord_format"],
            )
            if nbt is not None:
                entities_out.append(nbt)

        return entities_out



import pkgutil
import sys
import os

import amulet
from amulet import level
from amulet import operations

AMULET_PATH = amulet.__path__[0]

hiddenimports = (
    [name for _, name, _ in pkgutil.walk_packages(level.__path__, level.__name__ + ".")]
    + [
        name
        for _, name, _ in pkgutil.walk_packages(
            operations.__path__, operations.__name__ + "."
        )
    ]
    + ["amulet.api.structure"]
)

if sys.platform == "linux":
    binaries = [
        (
            os.path.join(
                AMULET_PATH, "libs", "leveldb", "leveldb_mcpe_linux_x86_64.so"
            ),
            os.path.join(".", "amulet", "libs", "leveldb"),
        ),
    ]
elif sys.platform == "win32":
    if sys.maxsize > 2**32:  # 64 bit python
        binaries = [
            (
                os.path.join(
                    AMULET_PATH, "libs", "leveldb", "leveldb_mcpe_win_amd64.dll"
                ),
                os.path.join(".", "amulet", "libs", "leveldb"),
            ),
        ]
    else:
        binaries = [
            (
                os.path.join(AMULET_PATH, "libs", "leveldb", "leveldb_mcpe_win32.dll"),
                os.path.join(".", "amulet", "libs", "leveldb"),
            )
        ]

elif sys.platform == "darwin":
    binaries = [
        (
            os.path.join(
                AMULET_PATH, "libs", "leveldb", "leveldb_mcpe_macosx_10_9_x86_64.dylib"
            ),
            os.path.join(".", "amulet", "libs", "leveldb"),
        ),
    ]
else:
    raise Exception(f"Unsupported platform {sys.platform}")



from __future__ import annotations

import importlib
from typing import AbstractSet, Any, Dict
import pkgutil
import logging
import inspect

from amulet.api.errors import LoaderNoneMatched
from amulet.api.wrapper import FormatWrapper, Interface, Translator

log = logging.getLogger(__name__)

ParentPackage = ".".join(__name__.split(".")[:-1])


class Loader:
    def __init__(
        self,
        base_class,
        object_type: str,
        package_name: str,
        create_instance=True,
    ):
        self._base_class = base_class
        self._object_type = object_type
        self._objects: Dict[str, Any] = {}
        self._create_instance = create_instance
        self._recursive_find(package_name)

    def _recursive_find(self, module_name: str):
        module = importlib.import_module(module_name)

        c = getattr(module, "export", None)
        if inspect.isclass(c) and issubclass(c, self._base_class):
            if self._create_instance:
                self._objects[module_name] = c()
            else:
                self._objects[module_name] = c

            log.debug(f'Enabled {self._object_type} "{module_name}"')

        elif hasattr(module, "__path__"):
            for _, sub_module_name, ispkg in pkgutil.iter_modules(
                module.__path__, module.__name__ + "."
            ):
                self._recursive_find(sub_module_name)

    def keys(self) -> AbstractSet[str]:
        """
        :return: The identifiers of all loaded objects
        """
        return self._objects.keys()

    def get(self, identifier: Any) -> Any:
        """
        Given an ``identifier`` will find a valid class and return it

        :param identifier: The identifier for the desired loaded object
        :return: The class for the object
        """
        object_id = self.identify(identifier)
        return self._objects[object_id]

    def identify(self, identifier: Any) -> str:
        if not self._objects:
            raise Exception(f"No {self._object_type} loaders found.")
        for object_name, obj in self._objects.items():
            if obj.is_valid(identifier):
                return object_name

        raise LoaderNoneMatched(
            f"Could not find a matching {self._object_type} for {identifier}"
        )

    def __contains__(self, item: str):
        return item in self._objects

    def report(self):
        print(f"{self._object_type} objects")
        for obj_name, obj in self._objects.items():
            print(obj_name, obj)


Translators = Loader(Translator, "translator", f"{ParentPackage}.translators")
Interfaces = Loader(Interface, "interface", f"{ParentPackage}.interfaces")
Formats = Loader(
    FormatWrapper, "format", f"{ParentPackage}.formats", create_instance=False
)


if __name__ == "__main__":
    Formats.report()
    Interfaces.report()
    Translators.report()



from .blocks import Blocks
from .status import Status, StatusFormats
from .biomes import Biomes, BiomesShape
from .entity_list import EntityList
from .block_entity_dict import BlockEntityDict
from .chunk import Chunk



from __future__ import annotations

from typing import TYPE_CHECKING

from amulet.api.data_types import Dimension, BlockCoordinates, FloatTriplet
from amulet.api.block import Block, UniversalAirLikeBlocks

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def paste(
    dst: "BaseLevel",
    dst_dimension: Dimension,
    src: "BaseLevel",
    src_dimension: Dimension,
    location: BlockCoordinates,
    scale: FloatTriplet,
    rotation: FloatTriplet,
    copy_air=True,
    copy_water=True,
    copy_lava=True,
):
    for _ in paste_iter(
        dst,
        dst_dimension,
        src,
        src_dimension,
        location,
        scale,
        rotation,
        copy_air,
        copy_water,
        copy_lava,
    ):
        pass


def paste_iter(
    dst: "BaseLevel",
    dst_dimension: Dimension,
    src: "BaseLevel",
    src_dimension: Dimension,
    location: BlockCoordinates,
    scale: FloatTriplet,
    rotation: FloatTriplet,
    copy_air=True,
    copy_water=True,
    copy_lava=True,
):
    yield from dst.paste_iter(
        src,
        src_dimension,
        src.bounds(src_dimension),
        dst_dimension,
        location,
        scale,
        rotation,
        True,
        True,
        tuple(UniversalAirLikeBlocks) * bool(not copy_air)
        + (Block("universal_minecraft", "water"),) * bool(not copy_water)
        + (Block("universal_minecraft", "lava"),) * bool(not copy_lava),
        True,
    )



from __future__ import annotations

import time
from typing import Union, Generator, Optional, Tuple, Callable, Set, Iterable
import traceback
import numpy
import itertools
import warnings
import logging
import copy
import os

from amulet.api.block import Block, UniversalAirBlock
from amulet.api.block_entity import BlockEntity
from amulet.api.entity import Entity
from amulet.api.registry import BlockManager
from amulet.api.registry.biome_manager import BiomeManager
from amulet.api.errors import ChunkDoesNotExist, ChunkLoadError, DimensionDoesNotExist
from amulet.api.chunk import Chunk, EntityList
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.data_types import (
    Dimension,
    VersionIdentifierType,
    BlockCoordinates,
    FloatTriplet,
    ChunkCoordinates,
)
from amulet.api.chunk.status import StatusFormats
from amulet.api.cache import TempDir
from amulet.libs.leveldb import LevelDB
from amulet.utils.generator import generator_unpacker
from amulet.utils.world_utils import block_coords_to_chunk_coords
from .chunk_manager import ChunkManager
from amulet.api.history.history_manager import MetaHistoryManager
from .clone import clone
from amulet.api import wrapper as api_wrapper, level as api_level
import PyMCTranslate
from amulet.api.player import Player
from .player_manager import PlayerManager

log = logging.getLogger(__name__)


class BaseLevel:
    """
    BaseLevel is a base class for all world-like data.

    It exposes chunk data and other data using a history system to track and enable undoing changes.
    """

    def __init__(self, path: str, format_wrapper: api_wrapper.FormatWrapper):
        """
        Construct a :class:`BaseLevel` object from the given data.

        This should not be used directly. You should instead use :func:`amulet.load_level`.

        :param path: The path to the data being loaded. May be a file or directory. If blank there is no data on disk associated with this.
        :param format_wrapper: The :class:`FormatWrapper` instance that the level will wrap around.
        """
        self._path = path

        self._level_wrapper = format_wrapper
        self.level_wrapper.open()

        self._block_palette = BlockManager()
        self._block_palette.get_add_block(
            UniversalAirBlock
        )  # ensure that index 0 is always air

        self._biome_palette = BiomeManager()
        self._biome_palette.get_add_biome("universal_minecraft:plains")

        self._history_manager = MetaHistoryManager()

        self._temp_dir = TempDir()
        self._history_db = LevelDB(
            os.path.join(self._temp_dir, "history_db"), create_if_missing=True
        )
        self._chunks: ChunkManager = ChunkManager(self, self._history_db)
        self._players = PlayerManager(self)

        self.history_manager.register(self._chunks, True)
        self.history_manager.register(self._players, True)

    def __del__(self):
        self.close()

    @property
    def level_wrapper(self) -> api_wrapper.FormatWrapper:
        """A class to access data directly from the level."""
        return self._level_wrapper

    @property
    def sub_chunk_size(self) -> int:
        """The normal dimensions of the chunk."""
        return self.level_wrapper.sub_chunk_size

    @property
    def level_path(self) -> str:
        """
        The system path where the level is located.

        This may be a directory, file or an empty string depending on the level that is loaded.
        """
        return self._path

    @property
    def translation_manager(self) -> PyMCTranslate.TranslationManager:
        """An instance of the translation class for use with this level."""
        return self.level_wrapper.translation_manager

    @property
    def block_palette(self) -> BlockManager:
        """The manager for the universal blocks in this level. New blocks must be registered here before adding to the level."""
        return self._block_palette

    @property
    def biome_palette(self) -> BiomeManager:
        """The manager for the universal blocks in this level. New biomes must be registered here before adding to the level."""
        return self._biome_palette

    @property
    def selection_bounds(self) -> SelectionGroup:
        """The selection(s) that all chunk data must fit within. Usually +/-30M for worlds. The selection for structures."""
        warnings.warn(
            "BaseLevel.selection_bounds is depreciated and will be removed in the future. Please use BaseLevel.bounds(dimension) instead",
            DeprecationWarning,
        )
        return self.bounds(self.dimensions[0])

    def bounds(self, dimension: Dimension) -> SelectionGroup:
        """
        The selection(s) that all chunk data must fit within.
        This specifies the volume that can be built in.
        Worlds will have a single cuboid volume.
        Structures may have one or more cuboid volumes.

        :param dimension: The dimension to get the bounds of.
        :return: The build volume for the dimension.
        """
        return self.level_wrapper.bounds(dimension)

    @property
    def dimensions(self) -> Tuple[Dimension, ...]:
        """The dimensions strings that are valid for this level."""
        return tuple(self.level_wrapper.dimensions)

    def get_block(self, x: int, y: int, z: int, dimension: Dimension) -> Block:
        """
        Gets the universal Block object at the specified coordinates.

        To get the block in a given format use :meth:`get_version_block`

        :param x: The X coordinate of the desired block
        :param y: The Y coordinate of the desired block
        :param z: The Z coordinate of the desired block
        :param dimension: The dimension of the desired block
        :return: The universal Block object representation of the block at that location
        :raises:
            :class:`~amulet.api.errors.ChunkDoesNotExist`: If the chunk does not exist (was deleted or never created)

            :class:`~amulet.api.errors.ChunkLoadError`: If the chunk was not able to be loaded. Eg. If the chunk is corrupt or some error occurred when loading.
        """
        cx, cz = block_coords_to_chunk_coords(x, z, sub_chunk_size=self.sub_chunk_size)
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        return self.get_chunk(cx, cz, dimension).get_block(offset_x, y, offset_z)

    def _chunk_box(self, cx: int, cz: int, sub_chunk_size: Optional[int] = None):
        """Get a SelectionBox containing the whole of a given chunk"""
        if sub_chunk_size is None:
            sub_chunk_size = self.sub_chunk_size
        return SelectionBox.create_chunk_box(cx, cz, sub_chunk_size)

    def _sanitise_selection(
        self, selection: Union[SelectionGroup, SelectionBox, None], dimension: Dimension
    ) -> SelectionGroup:
        if isinstance(selection, SelectionBox):
            return SelectionGroup(selection)
        elif isinstance(selection, SelectionGroup):
            return selection
        elif selection is None:
            return self.bounds(dimension)
        else:
            raise ValueError(
                f"Expected SelectionBox, SelectionGroup or None. Got {selection}"
            )

    def get_coord_box(
        self,
        dimension: Dimension,
        selection: Union[SelectionGroup, SelectionBox, None] = None,
        yield_missing_chunks=False,
    ) -> Generator[Tuple[ChunkCoordinates, SelectionBox], None, None]:
        """
        Given a selection will yield chunk coordinates and :class:`SelectionBox` instances into that chunk

        If not given a selection will use the bounds of the object.

        :param dimension: The dimension to take effect in.
        :param selection: SelectionGroup or SelectionBox into the level. If None will use :meth:`bounds` for the dimension.
        :param yield_missing_chunks: If a chunk does not exist an empty one will be created (defaults to false). Use this with care.
        """
        selection = self._sanitise_selection(selection, dimension)
        if yield_missing_chunks or selection.footprint_area < 1_000_000:
            if yield_missing_chunks:
                for coord, box in selection.chunk_boxes(self.sub_chunk_size):
                    yield coord, box
            else:
                for (cx, cz), box in selection.chunk_boxes(self.sub_chunk_size):
                    if self.has_chunk(cx, cz, dimension):
                        yield (cx, cz), box

        else:
            # if the selection gets very large iterating over the whole selection and accessing chunks can get slow
            # instead we are going to iterate over the chunks and get the intersection of the selection
            for cx, cz in self.all_chunk_coords(dimension):
                box = SelectionGroup(
                    SelectionBox.create_chunk_box(cx, cz, self.sub_chunk_size)
                )

                if selection.intersects(box):
                    chunk_selection = selection.intersection(box)
                    for sub_box in chunk_selection.selection_boxes:
                        yield (cx, cz), sub_box

    def get_chunk_boxes(
        self,
        dimension: Dimension,
        selection: Union[SelectionGroup, SelectionBox, None] = None,
        create_missing_chunks=False,
    ) -> Generator[Tuple[Chunk, SelectionBox], None, None]:
        """
        Given a selection will yield :class:`Chunk` and :class:`SelectionBox` instances into that chunk

        If not given a selection will use the bounds of the object.

        :param dimension: The dimension to take effect in.
        :param selection: SelectionGroup or SelectionBox into the level. If None will use :meth:`bounds` for the dimension.
        :param create_missing_chunks: If a chunk does not exist an empty one will be created (defaults to false). Use this with care.
        """
        for (cx, cz), box in self.get_coord_box(
            dimension, selection, create_missing_chunks
        ):
            try:
                chunk = self.get_chunk(cx, cz, dimension)
            except ChunkDoesNotExist:
                if create_missing_chunks:
                    yield self.create_chunk(cx, cz, dimension), box
            except ChunkLoadError:
                log.error(f"Error loading chunk\n{traceback.format_exc()}")
            else:
                yield chunk, box

    def get_chunk_slice_box(
        self,
        dimension: Dimension,
        selection: Union[SelectionGroup, SelectionBox] = None,
        create_missing_chunks=False,
    ) -> Generator[Tuple[Chunk, Tuple[slice, slice, slice], SelectionBox], None, None]:
        """
        Given a selection will yield :class:`Chunk`, slices, :class:`SelectionBox` for the contents of the selection.

        :param dimension: The dimension to take effect in.
        :param selection: SelectionGroup or SelectionBox into the level. If None will use :meth:`bounds` for the dimension.
        :param create_missing_chunks: If a chunk does not exist an empty one will be created (defaults to false)

        >>> for chunk, slices, box in level.get_chunk_slice_box(selection):
        >>>     chunk.blocks[slice] = ...
        """
        for chunk, box in self.get_chunk_boxes(
            dimension, selection, create_missing_chunks
        ):
            slices = box.chunk_slice(chunk.cx, chunk.cz, self.sub_chunk_size)
            yield chunk, slices, box

    def get_moved_coord_slice_box(
        self,
        dimension: Dimension,
        destination_origin: BlockCoordinates,
        selection: Optional[Union[SelectionGroup, SelectionBox]] = None,
        destination_sub_chunk_shape: Optional[int] = None,
        yield_missing_chunks: bool = False,
    ) -> Generator[
        Tuple[
            ChunkCoordinates,
            Tuple[slice, slice, slice],
            SelectionBox,
            ChunkCoordinates,
            Tuple[slice, slice, slice],
            SelectionBox,
        ],
        None,
        None,
    ]:
        """
        Iterate over a selection and return slices into the source object and destination object
        given the origin of the destination. When copying a selection to a new area the slices will
        only be equal if the offset is a multiple of the chunk size. This will rarely be the case
        so the slices need to be split up into parts that intersect a chunk in the source and destination.

        :param dimension: The dimension to iterate over.
        :param destination_origin: The location where the minimum point of the selection will end up
        :param selection: An optional selection. The overlap of this and the dimensions bounds will be used
        :param destination_sub_chunk_shape: the chunk shape of the destination object (defaults to self.sub_chunk_size)
        :param yield_missing_chunks: Generate empty chunks if the chunk does not exist.
        :return:
        """
        if destination_sub_chunk_shape is None:
            destination_sub_chunk_shape = self.sub_chunk_size

        if selection is None:
            selection = self.bounds(dimension)
        else:
            selection = self.bounds(dimension).intersection(selection)
        # the offset from self.selection to the destination location
        offset = numpy.subtract(
            destination_origin, self.bounds(dimension).min, dtype=int
        )
        for (src_cx, src_cz), box in self.get_coord_box(
            dimension, selection, yield_missing_chunks=yield_missing_chunks
        ):
            dst_full_box = SelectionBox(offset + box.min, offset + box.max)

            first_chunk = block_coords_to_chunk_coords(
                dst_full_box.min_x,
                dst_full_box.min_z,
                sub_chunk_size=destination_sub_chunk_shape,
            )
            last_chunk = block_coords_to_chunk_coords(
                dst_full_box.max_x - 1,
                dst_full_box.max_z - 1,
                sub_chunk_size=destination_sub_chunk_shape,
            )
            for dst_cx, dst_cz in itertools.product(
                range(first_chunk[0], last_chunk[0] + 1),
                range(first_chunk[1], last_chunk[1] + 1),
            ):
                chunk_box = self._chunk_box(dst_cx, dst_cz, destination_sub_chunk_shape)
                dst_box = chunk_box.intersection(dst_full_box)
                src_box = SelectionBox(-offset + dst_box.min, -offset + dst_box.max)
                src_slices = src_box.chunk_slice(src_cx, src_cz, self.sub_chunk_size)
                dst_slices = dst_box.chunk_slice(dst_cx, dst_cz, self.sub_chunk_size)
                yield (src_cx, src_cz), src_slices, src_box, (
                    dst_cx,
                    dst_cz,
                ), dst_slices, dst_box

    def get_moved_chunk_slice_box(
        self,
        dimension: Dimension,
        destination_origin: BlockCoordinates,
        selection: Optional[Union[SelectionGroup, SelectionBox]] = None,
        destination_sub_chunk_shape: Optional[int] = None,
        create_missing_chunks: bool = False,
    ) -> Generator[
        Tuple[
            Chunk,
            Tuple[slice, slice, slice],
            SelectionBox,
            ChunkCoordinates,
            Tuple[slice, slice, slice],
            SelectionBox,
        ],
        None,
        None,
    ]:
        """
        Iterate over a selection and return slices into the source object and destination object
        given the origin of the destination. When copying a selection to a new area the slices will
        only be equal if the offset is a multiple of the chunk size. This will rarely be the case
        so the slices need to be split up into parts that intersect a chunk in the source and destination.

        :param dimension: The dimension to iterate over.
        :param destination_origin: The location where the minimum point of self.selection will end up
        :param selection: An optional selection. The overlap of this and self.selection will be used
        :param destination_sub_chunk_shape: the chunk shape of the destination object (defaults to self.sub_chunk_size)
        :param create_missing_chunks: Generate empty chunks if the chunk does not exist.
        :return:
        """
        for (
            (src_cx, src_cz),
            src_slices,
            src_box,
            (dst_cx, dst_cz),
            dst_slices,
            dst_box,
        ) in self.get_moved_coord_slice_box(
            dimension,
            destination_origin,
            selection,
            destination_sub_chunk_shape,
            create_missing_chunks,
        ):
            try:
                chunk = self.get_chunk(src_cx, src_cz, dimension)
            except ChunkDoesNotExist:
                chunk = self.create_chunk(dst_cx, dst_cz, dimension)
            except ChunkLoadError:
                log.error(f"Error loading chunk\n{traceback.format_exc()}")
                continue
            yield chunk, src_slices, src_box, (dst_cx, dst_cz), dst_slices, dst_box

    def pre_save_operation(self) -> Generator[float, None, bool]:
        """
        Logic to run before saving. Eg recalculating height maps or lighting.
        Is a generator yielding progress from 0 to 1 and returning a bool saying if changes have been made.

        :return: Have any modifications been made.
        """
        return self.level_wrapper.pre_save_operation(self)

    def save(
        self,
        wrapper: api_wrapper.FormatWrapper = None,
        progress_callback: Callable[[int, int], None] = None,
    ):
        """
        Save the level to the given :class:`FormatWrapper`.

        :param wrapper: If specified will save the data to this wrapper instead of self.level_wrapper
        :param progress_callback: Optional progress callback to let the calling program know the progress. Input format chunk_index, chunk_count
        :return:
        """
        for chunk_index, chunk_count in self.save_iter(wrapper):
            if progress_callback is not None:
                progress_callback(chunk_index, chunk_count)

    def save_iter(
        self, wrapper: api_wrapper.FormatWrapper = None
    ) -> Generator[Tuple[int, int], None, None]:
        """
        Save the level to the given :class:`FormatWrapper`.

        This will yield the progress which can be used to update a UI.

        :param wrapper: If specified will save the data to this wrapper instead of self.level_wrapper
        :return: A generator of the number of chunks completed and the total number of chunks
        """
        # TODO change the yield type to match OperationReturnType

        chunk_index = 0

        changed_chunks = list(self._chunks.changed_chunks())
        chunk_count = len(changed_chunks)

        if wrapper is None:
            wrapper = self.level_wrapper

        output_dimension_map = wrapper.dimensions

        # perhaps make this check if the directory is the same rather than if the class is the same
        save_as = wrapper is not self.level_wrapper
        if save_as:
            # The input wrapper is not the same as the loading wrapper (save-as)
            # iterate through every chunk in the input level and save them to the wrapper
            log.info(
                f"Converting level {self.level_wrapper.path} to level {wrapper.path}"
            )
            wrapper.translation_manager = (
                self.level_wrapper.translation_manager
            )  # TODO: this might cause issues in the future
            for dimension in self.level_wrapper.dimensions:
                chunk_count += len(list(self.level_wrapper.all_chunk_coords(dimension)))

            for dimension in self.level_wrapper.dimensions:
                try:
                    if dimension not in output_dimension_map:
                        continue
                    for cx, cz in self.level_wrapper.all_chunk_coords(dimension):
                        log.info(f"Converting chunk {dimension} {cx}, {cz}")
                        try:
                            chunk = self.level_wrapper.load_chunk(cx, cz, dimension)
                            if chunk.status.as_type(StatusFormats.Java_14) == "full":
                                wrapper.commit_chunk(chunk, dimension)
                        except ChunkLoadError:
                            log.info(f"Error loading chunk {cx} {cz}", exc_info=True)
                        chunk_index += 1
                        yield chunk_index, chunk_count
                        if not chunk_index % 10000:
                            wrapper.save()
                            self.level_wrapper.unload()
                            wrapper.unload()
                except DimensionDoesNotExist:
                    continue

        for dimension, cx, cz in changed_chunks:
            if dimension not in output_dimension_map:
                continue
            try:
                chunk = self.get_chunk(cx, cz, dimension)
            except ChunkDoesNotExist:
                wrapper.delete_chunk(cx, cz, dimension)
            except ChunkLoadError:
                pass
            else:
                wrapper.commit_chunk(chunk, dimension)
                chunk.changed = False
            chunk_index += 1
            yield chunk_index, chunk_count
            if not chunk_index % 10000:
                wrapper.save()
                wrapper.unload()

        self.history_manager.mark_saved()
        log.info(f"Saving changes to level {wrapper.path}")
        wrapper.save()
        log.info(f"Finished saving changes to level {wrapper.path}")

    def purge(self):
        """
        Unload all loaded and cached data.

        This is functionally the same as closing and reopening the world without creating a new class.
        """
        self.unload()
        self.history_manager.purge()

    def close(self):
        """
        Close the attached level and remove temporary files.

        Use changed method to check if there are any changes that should be saved before closing.
        """
        self.level_wrapper.close()
        self._history_db.close(compact=False)

    def unload(self, safe_area: Optional[Tuple[Dimension, int, int, int, int]] = None):
        """
        Unload all chunk data not in the safe area.

        :param safe_area: The area that should not be unloaded [dimension, min_chunk_x, min_chunk_z, max_chunk_x, max_chunk_z]. If None will unload all chunk data.
        """
        self._chunks.unload(safe_area)
        self.level_wrapper.unload()

    def unload_unchanged(self):
        """Unload all data that has not been marked as changed."""
        self._chunks.unload_unchanged()

    @property
    def chunks(self) -> ChunkManager:
        """
        The chunk container.

        Most methods from :class:`ChunkManager` also exists in the level class.
        """
        return self._chunks

    def all_chunk_coords(self, dimension: Dimension) -> Set[Tuple[int, int]]:
        """
        The coordinates of every chunk in this dimension of the level.

        This is the combination of chunks saved to the level and chunks yet to be saved.
        """
        return self._chunks.all_chunk_coords(dimension)

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        """
        Does the chunk exist. This is a quick way to check if the chunk exists without loading it.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the chunk from.
        :return: True if the chunk exists. Calling get_chunk on this chunk may still throw ChunkLoadError
        """
        return self._chunks.has_chunk(dimension, cx, cz)

    def get_chunk(self, cx: int, cz: int, dimension: Dimension) -> Chunk:
        """
        Gets a :class:`Chunk` class containing the data for the requested chunk.

        :param cx: The X coordinate of the desired chunk
        :param cz: The Z coordinate of the desired chunk
        :param dimension: The dimension to get the chunk from
        :return: A Chunk object containing the data for the chunk
        :raises:
            :class:`~amulet.api.errors.ChunkDoesNotExist`: If the chunk does not exist (was deleted or never created)

            :class:`~amulet.api.errors.ChunkLoadError`: If the chunk was not able to be loaded. Eg. If the chunk is corrupt or some error occurred when loading.
        """
        return self._chunks.get_chunk(dimension, cx, cz)

    def create_chunk(self, cx: int, cz: int, dimension: Dimension) -> Chunk:
        """
        Create an empty chunk and put it at the given location.

        If a chunk exists at the given location it will be overwritten.

        :param cx: The X coordinate of the chunk
        :param cz: The Z coordinate of the chunk
        :param dimension: The dimension to put the chunk in.
        :return: The newly created :class:`Chunk`.
        """
        chunk = Chunk(cx, cz)
        self.put_chunk(chunk, dimension)
        return chunk

    def put_chunk(self, chunk: Chunk, dimension: Dimension):
        """
        Add a given chunk to the level.

        :param chunk: The :class:`Chunk` to add to the level. It will be added at the location stored in :attr:`Chunk.coordinates`
        :param dimension: The dimension to add the chunk to.
        """
        self._chunks.put_chunk(chunk, dimension)

    def delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        """
        Delete a chunk from the level.

        :param cx: The X coordinate of the chunk
        :param cz: The Z coordinate of the chunk
        :param dimension: The dimension to delete the chunk from.
        """
        self._chunks.delete_chunk(dimension, cx, cz)

    def extract_structure(
        self, selection: SelectionGroup, dimension: Dimension
    ) -> api_level.ImmutableStructure:
        """
        Extract the region of the dimension specified by ``selection`` to an :class:`~api_level.ImmutableStructure` class.

        :param selection: The selection to extract.
        :param dimension: The dimension to extract the selection from.
        :return: The :class:`~api_level.ImmutableStructure` containing the extracted region.
        """
        return api_level.ImmutableStructure.from_level(self, selection, dimension)

    def extract_structure_iter(
        self, selection: SelectionGroup, dimension: Dimension
    ) -> Generator[float, None, api_level.ImmutableStructure]:
        """
        Extract the region of the dimension specified by ``selection`` to an :class:`~api_level.ImmutableStructure` class.

        Also yields the progress as a float from 0-1

        :param selection: The selection to extract.
        :param dimension: The dimension to extract the selection from.
        :return: The :class:`~api_level.ImmutableStructure` containing the extracted region.
        """
        immutable_level = yield from api_level.ImmutableStructure.from_level_iter(
            self, selection, dimension
        )
        return immutable_level

    def paste(
        self,
        src_structure: "BaseLevel",
        src_dimension: Dimension,
        src_selection: SelectionGroup,
        dst_dimension: Dimension,
        location: BlockCoordinates,
        scale: FloatTriplet = (1.0, 1.0, 1.0),
        rotation: FloatTriplet = (0.0, 0.0, 0.0),
        include_blocks: bool = True,
        include_entities: bool = True,
        skip_blocks: Tuple[Block, ...] = (),
        copy_chunk_not_exist: bool = False,
    ):
        """Paste a level into this level at the given location.
        Note this command may change in the future.

        :param src_structure: The structure to paste into this structure.
        :param src_dimension: The dimension of the source structure to copy from.
        :param src_selection: The selection to copy from the source structure.
        :param dst_dimension: The dimension to paste the structure into.
        :param location: The location where the centre of the structure will be in the level
        :param scale: The scale in the x, y and z axis. These can be negative to mirror.
        :param rotation: The rotation in degrees around each of the axis.
        :param include_blocks: Include blocks when pasting the structure.
        :param include_entities: Include entities when pasting the structure.
        :param skip_blocks: If a block matches a block in this list it will not be copied.
        :param copy_chunk_not_exist: If a chunk does not exist in the source should it be copied over as air. Always False where level is a World.
        :return:
        """
        return generator_unpacker(
            self.paste_iter(
                src_structure,
                src_dimension,
                src_selection,
                dst_dimension,
                location,
                scale,
                rotation,
                include_blocks,
                include_entities,
                skip_blocks,
                copy_chunk_not_exist,
            )
        )

    def paste_iter(
        self,
        src_structure: "BaseLevel",
        src_dimension: Dimension,
        src_selection: SelectionGroup,
        dst_dimension: Dimension,
        location: BlockCoordinates,
        scale: FloatTriplet = (1.0, 1.0, 1.0),
        rotation: FloatTriplet = (0.0, 0.0, 0.0),
        include_blocks: bool = True,
        include_entities: bool = True,
        skip_blocks: Tuple[Block, ...] = (),
        copy_chunk_not_exist: bool = False,
    ) -> Generator[float, None, None]:
        """Paste a structure into this structure at the given location.
        Note this command may change in the future.

        :param src_structure: The structure to paste into this structure.
        :param src_dimension: The dimension of the source structure to copy from.
        :param src_selection: The selection to copy from the source structure.
        :param dst_dimension: The dimension to paste the structure into.
        :param location: The location where the centre of the structure will be in the level
        :param scale: The scale in the x, y and z axis. These can be negative to mirror.
        :param rotation: The rotation in degrees around each of the axis.
        :param include_blocks: Include blocks when pasting the structure.
        :param include_entities: Include entities when pasting the structure.
        :param skip_blocks: If a block matches a block in this list it will not be copied.
        :param copy_chunk_not_exist: If a chunk does not exist in the source should it be copied over as air. Always False where level is a World.
        :return: A generator of floats from 0 to 1 with the progress of the paste operation.
        """
        yield from clone(
            src_structure,
            src_dimension,
            src_selection,
            self,
            dst_dimension,
            self.bounds(dst_dimension),
            location,
            scale,
            rotation,
            include_blocks,
            include_entities,
            skip_blocks,
            copy_chunk_not_exist,
        )

    def get_version_block(
        self,
        x: int,
        y: int,
        z: int,
        dimension: Dimension,
        version: VersionIdentifierType,
    ) -> Union[Tuple[Block, BlockEntity], Tuple[Entity, None]]:
        """
        Get a block at the specified location and convert it to the format of the version specified

        Note the odd return format. In most cases this will return (Block, None) or (Block, BlockEntity) if a block entity is present.

        In select cases (like item frames) it may return (Entity, None)

        :param x: The X coordinate of the desired block
        :param y: The Y coordinate of the desired block
        :param z: The Z coordinate of the desired block
        :param dimension: The dimension of the desired block
        :param version: The version to get the block converted to.

            >>> ("java", (1, 16, 2))  # Java 1.16.2 format
            >>> ("java", 2578)  # Java 1.16.2 format (using the data version)
            >>> ("bedrock", (1, 16, 210))  # Bedrock 1.16.210 format
        :return: The block at the given location converted to the `version` format. Note the odd return format.
        :raises:
            :class:`~amulet.api.errors.ChunkDoesNotExist`: If the chunk does not exist (was deleted or never created)

            :class:`~amulet.api.errors.ChunkLoadError`: If the chunk was not able to be loaded. Eg. If the chunk is corrupt or some error occurred when loading.
        """
        cx, cz = block_coords_to_chunk_coords(x, z, sub_chunk_size=self.sub_chunk_size)
        chunk = self.get_chunk(cx, cz, dimension)
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        output, extra_output, _ = self.translation_manager.get_version(
            *version
        ).block.from_universal(
            chunk.get_block(offset_x, y, offset_z), chunk.block_entities.get((x, y, z))
        )
        return output, extra_output

    def set_version_block(
        self,
        x: int,
        y: int,
        z: int,
        dimension: Dimension,
        version: VersionIdentifierType,
        block: Block,
        block_entity: BlockEntity = None,
    ):
        """
        Convert the block and block_entity from the given version format to the universal format and set at the location.

        :param x: The X coordinate of the desired block.
        :param y: The Y coordinate of the desired block.
        :param z: The Z coordinate of the desired block.
        :param dimension: The dimension of the desired block.
        :param version: The version the given ``block`` and ``block_entity`` come from.

            >>> ("java", (1, 16, 2))  # Java 1.16.2 format
            >>> ("java", 2578)  # Java 1.16.2 format (using the data version)
            >>> ("bedrock", (1, 16, 210))  # Bedrock 1.16.210 format
        :param block: The block to set. Must be valid in the specified version.
        :param block_entity: The block entity to set. Must be valid in the specified version.
        :return: The block at the given location converted to the `version` format. Note the odd return format.
        :raises:
            ChunkLoadError: If the chunk was not able to be loaded. Eg. If the chunk is corrupt or some error occurred when loading.
        """
        cx, cz = block_coords_to_chunk_coords(x, z, sub_chunk_size=self.sub_chunk_size)
        try:
            chunk = self.get_chunk(cx, cz, dimension)
        except ChunkDoesNotExist:
            chunk = self.create_chunk(cx, cz, dimension)
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        (
            universal_block,
            universal_block_entity,
            _,
        ) = self.translation_manager.get_version(*version).block.to_universal(
            block, block_entity
        )
        chunk.set_block(offset_x, y, offset_z, universal_block),
        if isinstance(universal_block_entity, BlockEntity):
            chunk.block_entities[(x, y, z)] = universal_block_entity
        elif (x, y, z) in chunk.block_entities:
            del chunk.block_entities[(x, y, z)]
        chunk.changed = True

    def get_native_entities(
        self, cx: int, cz: int, dimension: Dimension
    ) -> Tuple[EntityList, VersionIdentifierType]:
        """
        Get a list of entities in the native format from a given chunk.
        This currently returns the raw data from the chunk but in the future will convert to the world version format.

        :param cx: The chunk x position
        :param cz: The chunk z position
        :param dimension: The dimension of the chunk.
        :return: A copy of the list of entities and the version format they are in.
        """
        chunk = self.get_chunk(cx, cz, dimension)
        # To make this forwards compatible this needs to be deep copied
        return copy.deepcopy(chunk._native_entities), chunk._native_version

    def set_native_entites(
        self, cx: int, cz: int, dimension: Dimension, entities: Iterable[Entity]
    ):
        """
        Set the entities in the native format.
        Note that the format must be compatible with `level_wrapper.max_world_version`.

        :param cx: The chunk x position
        :param cz: The chunk z position
        :param dimension: The dimension of the chunk.
        :param entities: The entities to set on the chunk.
        """
        chunk = self.get_chunk(cx, cz, dimension)
        chunk._native_entities = EntityList(copy.deepcopy(entities))
        chunk._native_version = self.level_wrapper.max_world_version
        chunk.changed = True

    # def get_entities_in_box(
    #     self, box: "SelectionGroup"
    # ) -> Generator[Tuple[Coordinates, List[object]], None, None]:
    #     # TODO: some of this logic can probably be moved the chunk class and have this method call that
    #     # TODO: update this to use the newer entity API
    #     out_of_place_entities = []
    #     entity_map: Dict[Tuple[int, int], List[List[object]]] = {}
    #     for chunk, subbox in self.get_chunk_boxes(box):
    #         entities = chunk.entities
    #         in_box = list(filter(lambda e: e.location in subbox, entities))
    #         not_in_box = filter(lambda e: e.location not in subbox, entities)
    #
    #         in_box_copy = deepcopy(in_box)
    #
    #         entity_map[chunk.coordinates] = [
    #             not_in_box,
    #             in_box,
    #         ]  # First index is the list of entities not in the box, the second is for ones that are
    #
    #         yield chunk.coordinates, in_box_copy
    #
    #         if (
    #             in_box != in_box_copy
    #         ):  # If an entity has been changed, update the dictionary entry
    #             entity_map[chunk.coordinates][1] = in_box_copy
    #         else:  # Delete the entry otherwise
    #             del entity_map[chunk.coordinates]
    #
    #     for chunk_coords, entity_list_list in entity_map.items():
    #         chunk = self.get_chunk(*chunk_coords)
    #         in_place_entities = list(
    #             filter(
    #                 lambda e: chunk_coords
    #                 == entity_position_to_chunk_coordinates(e.location),
    #                 entity_list_list[1],
    #             )
    #         )
    #         out_of_place = filter(
    #             lambda e: chunk_coords
    #             != entity_position_to_chunk_coordinates(e.location),
    #             entity_list_list[1],
    #         )
    #
    #         chunk.entities = in_place_entities + list(entity_list_list[0])
    #
    #         if out_of_place:
    #             out_of_place_entities.extend(out_of_place)
    #
    #     if out_of_place_entities:
    #         self.add_entities(out_of_place_entities)
    #
    # def add_entities(self, entities):
    #     proper_entity_chunks = map(
    #         lambda e: (entity_position_to_chunk_coordinates(e.location), e,), entities,
    #     )
    #     accumulated_entities: Dict[Tuple[int, int], List[object]] = {}
    #
    #     for chunk_coord, ent in proper_entity_chunks:
    #         if chunk_coord in accumulated_entities:
    #             accumulated_entities[chunk_coord].append(ent)
    #         else:
    #             accumulated_entities[chunk_coord] = [ent]
    #
    #     for chunk_coord, ents in accumulated_entities.items():
    #         chunk = self.get_chunk(*chunk_coord)
    #
    #         chunk.entities += ents
    #
    # def delete_entities(self, entities):
    #     chunk_entity_pairs = map(
    #         lambda e: (entity_position_to_chunk_coordinates(e.location), e,), entities,
    #     )
    #
    #     for chunk_coord, ent in chunk_entity_pairs:
    #         chunk = self.get_chunk(*chunk_coord)
    #         entities = chunk.entities
    #         entities.remove(ent)
    #         chunk.entities = entities

    @property
    def history_manager(self) -> MetaHistoryManager:
        """The class that manages undoing and redoing changes."""
        return self._history_manager

    def create_undo_point(self, world=True, non_world=True) -> bool:
        """
        Create a restore point for all the data that has changed.

        :param world: If True the restore point will include world based data.
        :param non_world: If True the restore point will include data not related to the world.
        :return: If True a restore point was created. If nothing changed no restore point will be created.
        """
        return self.history_manager.create_undo_point(world, non_world)

    def create_undo_point_iter(
        self, world=True, non_world=True
    ) -> Generator[float, None, bool]:
        """
        Create a restore point for all the data that has changed.

        Also yields progress from 0-1

        :param world: If True the restore point will include world based data.
        :param non_world: If True the restore point will include data not related to the world.
        :return: If True a restore point was created. If nothing changed no restore point will be created.
        """
        return self.history_manager.create_undo_point_iter(world, non_world)

    @property
    def changed(self) -> bool:
        """Has any data been modified but not saved to disk"""
        return self.history_manager.changed or self.level_wrapper.changed

    def undo(self):
        """Undoes the last set of changes to the level."""
        self.history_manager.undo()

    def redo(self):
        """Redoes the last set of changes to the level."""
        self.history_manager.redo()

    def restore_last_undo_point(self):
        """
        Restore the level to the state it was when self.create_undo_point was last called.

        If an operation errors there may be modifications made that did not get tracked.

        This will revert those changes.
        """
        self.history_manager.restore_last_undo_point()

    @property
    def players(self) -> PlayerManager:
        """
        The player container.

        Most methods from :class:`PlayerManager` also exists in the level class.
        """
        return self._players

    def all_player_ids(self) -> Set[str]:
        """
        Returns a set of all player ids that are present in the level.
        """
        return self.players.all_player_ids()

    def has_player(self, player_id: str) -> bool:
        """
        Is the given player id present in the level

        :param player_id: The player id to check
        :return: True if the player id is present, False otherwise
        """
        return self.players.has_player(player_id)

    def get_player(self, player_id: str) -> Player:
        """
        Gets the :class:`Player` object that belongs to the specified player id

        If no parameter is supplied, the data of the local player will be returned

        :param player_id: The desired player id
        :return: A Player instance
        """
        return self.players.get_player(player_id)



from __future__ import annotations

from .anvil_2203 import Anvil2203Interface as ParentInterface


class Anvil2529Interface(ParentInterface):
    """
    Packed long arrays switched to a less dense format
    Before the long array was just a bit stream but it is now separate longs. The upper bits are unused in some cases.
    """

    LongArrayDense = False

    @staticmethod
    def minor_is_valid(key: int):
        return 2529 <= key < 2681


export = Anvil2529Interface



from typing import Any, List, Dict, Tuple, Optional, TYPE_CHECKING, Iterable, Union

from amulet.api.data_types import (
    Dimension,
    PlatformType,
    ChunkCoordinates,
    AnyNDArray,
    VersionNumberTuple,
)
from amulet.api.wrapper import FormatWrapper
from amulet.api.errors import ChunkDoesNotExist, PlayerDoesNotExist
from amulet.api.player import Player
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup
from amulet.api import wrapper as api_wrapper

if TYPE_CHECKING:
    from amulet.api.wrapper import Interface


class VoidFormatWrapper(FormatWrapper[VersionNumberTuple]):
    """
    A custom :class:`FormatWrapper` class that has no associated data.

    This is just to make the :class:`ImmutableStructure` class happy since it requires a :class:`FormatWrapper` class.

    All methods effectively do nothing.
    """

    def __init__(self, path: str):
        super().__init__(path)
        self._platform = "Unknown Platform"
        self._version = (0, 0, 0)

    @property
    def level_name(self) -> str:
        return "Void"

    @staticmethod
    def is_valid(path: str) -> bool:
        return False

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {}

    @property
    def dimensions(self) -> List[Dimension]:
        return ["main"]

    @property
    def can_add_dimension(self) -> bool:
        return False

    def register_dimension(self, dimension_identifier: Any):
        pass

    def _get_interface(self, raw_chunk_data: Optional[Any] = None) -> "Interface":
        raise Exception("If this is called something is wrong")

    def _encode(
        self,
        interface: api_wrapper.Interface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ) -> Any:
        raise Exception("If this is called something is wrong")

    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        **kwargs
    ):
        pass

    def _open(self):
        pass

    def _save(self):
        pass

    def _close(self):
        pass

    def unload(self):
        pass

    def all_chunk_coords(self, dimension: Dimension) -> Iterable[ChunkCoordinates]:
        yield from ()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return False

    def _delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        pass

    def _put_raw_chunk_data(self, cx: int, cz: int, data: Any, dimension: Dimension):
        pass

    def _get_raw_chunk_data(self, cx: int, cz: int, dimension: Dimension) -> Any:
        raise ChunkDoesNotExist

    def all_player_ids(self) -> Iterable[str]:
        yield from ()

    def has_player(self, player_id: str) -> bool:
        return False

    def _load_player(self, player_id: str) -> Player:
        raise PlayerDoesNotExist

    def _get_raw_player_data(self, player_id: str) -> Any:
        raise PlayerDoesNotExist



from __future__ import annotations

import math
import sys
import gzip
from io import StringIO
from typing import Tuple, Optional
import numpy
from numpy import ndarray, zeros, uint8
from amulet.api.data_types import ChunkCoordinates

SECTOR_BYTES = 4096
SECTOR_INTS = SECTOR_BYTES / 4
CHUNK_HEADER_SIZE = 5
VERSION_GZIP = 1
VERSION_DEFLATE = 2


def block_coords_to_chunk_coords(
    *args: int, sub_chunk_size: int = 16
) -> Tuple[int, ...]:
    """
    Converts the supplied block coordinates into chunk coordinates

    :param args: The coordinate of the block(s)
    :param sub_chunk_size: The dimension of the chunk (Optional. Default 16)
    :return: The resulting chunk coordinates in (x, z) order
    """
    return tuple(int(math.floor(coord / sub_chunk_size)) for coord in args)


def chunk_coords_to_block_coords(
    x: int, z: int, chunk_x_size: int = 16, chunk_z_size: int = 16
) -> ChunkCoordinates:
    """
    Converts the supplied chunk coordinates into block coordinates

    :param x: The x coordinate of the chunk
    :param z: The z coordinate of the chunk
    :param chunk_x_size: The dimension of the chunk in the x direction (Optional. Default 16)
    :param chunk_z_size: The dimension of the chunk in the z direction (Optional. Default 16)
    :return: The resulting block coordinates in (x, z) order
    """
    return x * chunk_x_size, z * chunk_z_size


def chunk_coords_to_region_coords(cx: int, cz: int) -> ChunkCoordinates:
    """
    Converts the supplied chunk coordinates into region coordinates

    :param cx: The x coordinate of the chunk
    :param cz: The z coordinate of the chunk
    :return: The resulting region coordinates in (x, z) order
    """
    return cx >> 5, cz >> 5


def region_coords_to_chunk_coords(rx: int, rz: int) -> ChunkCoordinates:
    """
    Converts the supplied region coordinates into the minimum chunk coordinates of that region

    :param rx: The x coordinate of the region
    :param rz: The y coordinate of the region
    :return: The resulting minimum chunk coordinates of that region in (x, z) order
    """
    return rx << 5, rz << 5


def blocks_slice_to_chunk_slice(
    blocks_slice: slice, chunk_shape: int, chunk_coord: int
) -> slice:
    """
    Converts the supplied blocks slice into chunk slice

    :param blocks_slice: The slice of the blocks
    :param chunk_shape: The shape of the chunk in this direction
    :param chunk_coord: The coordinate of the chunk in this direction
    :return: The resulting chunk slice
    """
    return slice(
        min(max(0, blocks_slice.start - chunk_coord * chunk_shape), chunk_shape),
        min(max(0, blocks_slice.stop - chunk_coord * chunk_shape), chunk_shape),
    )


def gunzip(data):
    """
    Decompresses data that is in Gzip format
    """
    return gzip.GzipFile(fileobj=StringIO(data)).read()


def from_nibble_array(arr: ndarray) -> ndarray:
    """
    Unpacks a flat nibble array into a full size numpy array

    :param arr: The nibble array
    :return: The resulting array
    """
    shape = arr.size

    new_arr = zeros((shape * 2), dtype=uint8)

    new_arr[::2] = arr & 0xF
    new_arr[1::2] = arr >> 4

    return new_arr


def to_nibble_array(arr: ndarray) -> ndarray:
    """
    packs a full size numpy array into a nibble array.

    :param arr: Full numpy array
    :return: The nibble array
    """
    arr = arr.ravel()
    return (arr[::2] + (arr[1::2] << 4)).astype("uint8")


"""
Minecraft Java edition stores the block and height arrays in a compacted long array format.
The format stores one or more entries per long, using the fewest number of bits required to store the data.
There are two storage methods, the compact version was used prior to 1.16 and the less compact version in 1.16 and above.
Apparently the less compact version is quicker to pack and unpack.
The compact version effectively stores the values as a bit array spanning one or more values in the long array.
There may be some padding if the bit array does not fill all the long values. (The letter "P" signifies an unused padding bit)
PPAAAAAAAAABBBBBBBBBCCCCCCCCCDDDDDDDDDEEEEEEEEEFFFFFFFFFGGGGGGGG GHHHHHHHHHIIIIIIIIIJJJJJJJJJKKKKKKKKKLLLLLLLLLMMMMMMMMMOOOOOOOOO
The less compact version does not allow entries to straddle long values. Instead, if required, there is padding within each long.
PAAAAAAAAABBBBBBBBBCCCCCCCCCDDDDDDDDDEEEEEEEEEFFFFFFFFFGGGGGGGGG PHHHHHHHHHIIIIIIIIIJJJJJJJJJKKKKKKKKKLLLLLLLLLMMMMMMMMMOOOOOOOOO

The functions below can be used to pack and unpack both formats.
"""


def decode_long_array(
    long_array: numpy.ndarray,
    size: int,
    bits_per_entry: int,
    dense=True,
    signed: bool = False,
) -> numpy.ndarray:
    """
    Decode a long array (from BlockStates or Heightmaps)

    :param long_array: Encoded long array
    :param size: The expected size of the returned array
    :param bits_per_entry: The number of bits per entry in the encoded array.
    :param dense: If true the long arrays will be treated as a bit stream. If false they are distinct values with padding
    :param signed: Should the returned array be signed.
    :return: Decoded array as numpy array
    """
    # validate the inputs and throw an error if there is a problem
    if not isinstance(bits_per_entry, int):
        raise ValueError(f"The bits_per_entry input must be an int.")

    assert (
        1 <= bits_per_entry <= 64
    ), f"bits_per_entry must be between 1 and 64 inclusive. Got {bits_per_entry}"

    # force the array to be a signed long array
    long_array = long_array.astype(">q")

    if dense:
        expected_len = math.ceil(size * bits_per_entry / 64)
    else:
        expected_len = math.ceil(size / (64 // bits_per_entry))
    if len(long_array) != expected_len:
        raise Exception(
            f"{'Dense e' if dense else 'E'}ncoded long array with {bits_per_entry} bits per entry should contain {expected_len} longs but got {len(long_array)}."
        )

    # unpack the long array into a bit array
    bits = numpy.unpackbits(long_array[::-1].astype(">i8").view("uint8"))
    if dense:
        if bits.size % bits_per_entry:
            # if the array is densely packed and there is extra padding, remove it
            bits = bits[bits.size % bits_per_entry :]
    else:
        # if not densely packed remove the padding per long
        entry_per_long = 64 // bits_per_entry
        bits = bits.reshape(-1, 64)[:, -entry_per_long * bits_per_entry :]

    byte_length = 2 ** math.ceil(math.log(math.ceil(bits_per_entry / 8), 2))
    dtype = {1: "B", 2: ">H", 4: ">I", 8: ">Q"}[byte_length]

    # pad the bits to fill one of the above data types
    arr = numpy.packbits(
        numpy.pad(
            bits.reshape(-1, bits_per_entry)[-size:, :],
            [(0, 0), (byte_length * 8 - bits_per_entry, 0)],
            "constant",
        )
    ).view(dtype=dtype)[::-1]
    if signed:
        # convert to a signed array if requested
        sarray = arr.astype({1: "b", 2: ">h", 4: ">i", 8: ">q"}[byte_length])
        if bits_per_entry < 64:
            mask = arr >= 2 ** (bits_per_entry - 1)
            sarray[mask] = numpy.subtract(
                arr[mask], 2**bits_per_entry, dtype=numpy.int64
            )
        arr = sarray
    return arr


def encode_long_array(
    array: numpy.ndarray,
    bits_per_entry: Optional[int] = None,
    dense: bool = True,
    min_bits_per_entry=1,
) -> numpy.ndarray:
    """
    Encode a long array (from BlockStates or Heightmaps)

    :param array: A numpy array of the data to be encoded.
    :param bits_per_entry: The number of bits to use to store each value. If left as None will use the smallest bits per entry.
    :param dense: If true the long arrays will be treated as a bit stream. If false they are distinct values with padding
    :param min_bits_per_entry: The mimimum value that bits_per_entry can be. If it is less than this it will be capped at this value.
    :return: Encoded array as numpy array
    """
    assert (
        1 <= min_bits_per_entry <= 64
    ), f"min_bits_per_entry must be between 1 and 64 inclusive. Got {bits_per_entry}"
    # cast to a signed longlong array
    array = array.astype(">q")
    # work out how many bits are required to store the
    required_bits_per_entry = max(
        max(
            int(numpy.amin(array)).bit_length(),
            int(numpy.amax(array)).bit_length(),
        ),
        min_bits_per_entry,
    )
    if bits_per_entry is None:
        # if a bit depth has not been requested use the minimum required
        bits_per_entry = required_bits_per_entry
    elif isinstance(bits_per_entry, int):
        assert (
            1 <= bits_per_entry <= 64
        ), f"bits_per_entry must be between 1 and 64 inclusive. Got {bits_per_entry}"
        # if a bit depth has been set and it is smaller than what is required throw an error
        if required_bits_per_entry > bits_per_entry:
            raise Exception(
                f"The array requires at least {required_bits_per_entry} bits per value which is more than the specified {bits_per_entry} bits"
            )
    else:
        raise ValueError(
            "bits_per_entry must be an int between 1 and 64 inclusive or None."
        )
    # make the negative values positive to make bit storage easier
    uarray = array.astype(">Q")
    if bits_per_entry < 64:
        mask = array < 0
        uarray[mask] = numpy.add(
            array[mask], 2**bits_per_entry, dtype=numpy.uint64, casting="unsafe"
        )
    array = uarray

    # unpack the individual values into a bit array
    bits: numpy.ndarray = numpy.unpackbits(
        numpy.ascontiguousarray(array[::-1]).view("uint8")
    ).reshape(-1, 64)[:, -bits_per_entry:]
    if dense:
        if bits.size % 64:
            # if the bit array does not fill a whole long
            # add padding to the last long if required
            bits = numpy.pad(
                bits.ravel(),
                [(64 - (bits.size % 64), 0)],
                "constant",
            )
    else:
        # if the array is not dense add padding
        entry_per_long = 64 // bits_per_entry
        if bits.shape[0] % entry_per_long:
            # add padding to the last long if required
            bits = numpy.pad(
                bits,
                [(entry_per_long - (bits.shape[0] % entry_per_long), 0), (0, 0)],
                "constant",
            )
        # add padding for each long
        bits = numpy.pad(
            bits.reshape(-1, bits_per_entry * entry_per_long),
            [(0, 0), (64 - bits_per_entry * entry_per_long, 0)],
            "constant",
        )

    # pack the bits into a long array
    return numpy.packbits(bits).view(dtype=">q")[::-1]


def get_size(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0

    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, "__dict__"):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size


def get_smallest_dtype(arr: ndarray, uint: bool = True) -> int:
    """
    Returns the smallest dtype (number) that the array can afford

    :param arr: The array to check on
    :param uint: Should the array fit in uint or not (default: True)
    :return: The number of bits all the elements can be represented with
    """
    possible_dtypes = (2**x for x in range(3, 8))
    max_number = numpy.amax(arr)
    if not uint:
        max_number = max_number * 2
    if max_number == 0:
        max_number = 1
    return next(dtype for dtype in possible_dtypes if dtype > math.log(max_number, 2))


def entity_position_to_chunk_coordinates(
    entity_coordinates: Tuple[float, float, float]
):
    return (
        int(math.floor(entity_coordinates[0])) >> 4,
        int(math.floor(entity_coordinates[2])) >> 4,
    )


def fast_unique(array: numpy.ndarray) -> Tuple[numpy.ndarray, numpy.ndarray]:
    uni = numpy.unique(array)
    lut = numpy.zeros(numpy.amax(uni) + 1, dtype=numpy.uint32)
    lut[uni] = numpy.arange(uni.size)
    inv = lut[array]
    return uni, inv



from abc import abstractmethod
import os
from typing import Any, Optional, TYPE_CHECKING

from amulet import IMG_DIRECTORY
from .format_wrapper import FormatWrapper, VersionNumberT

if TYPE_CHECKING:
    from amulet.api.wrapper import Interface

missing_world_icon = os.path.abspath(
    os.path.join(IMG_DIRECTORY, "missing_world_icon.png")
)


class WorldFormatWrapper(FormatWrapper[VersionNumberT]):
    _missing_world_icon = missing_world_icon

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`WorldFormatWrapper`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        super().__init__(path)
        self._world_image_path = missing_world_icon

    @property
    @abstractmethod
    def level_name(self) -> str:
        raise NotImplementedError

    @level_name.setter
    @abstractmethod
    def level_name(self, value: str):
        raise NotImplementedError

    @property
    @abstractmethod
    def last_played(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def game_version_string(self) -> str:
        raise NotImplementedError

    @property
    def world_image_path(self) -> str:
        """The path to the world icon."""
        return self._world_image_path

    @property
    def can_add_dimension(self) -> bool:
        return False

    def register_dimension(self, dimension_identifier: Any):
        pass

    def _get_interface(self, raw_chunk_data: Optional[Any] = None) -> "Interface":
        raise NotImplementedError



import argparse
import os

parser = argparse.ArgumentParser(
    description="Generate all interfaces up to a given level."
)
parser.add_argument(
    "interfaceversion", type=int, help="Generate all interfaces up to this level."
)


if __name__ == "__main__":
    args = parser.parse_args()
    for i in range(1, args.interfaceversion):
        path = os.path.join(os.path.dirname(__file__), f"leveldb_{i}.py")
        if not os.path.isfile(path):
            with open(path, "w") as f:
                f.write(
                    f"""from .leveldb_{i-1} import (
    LevelDB{i-1}Interface as ParentInterface,
)


class LevelDB{i}Interface(ParentInterface):
    chunk_version = {i}


export = LevelDB{i}Interface

"""
                )



from __future__ import annotations

from abc import ABC
from typing import Union, Optional
from amulet_nbt import NamedTag, CompoundTag
from amulet.api.data_types import BlockCoordinates, PointCoordinates

_Coord = Union[float, int]


class AbstractBaseEntity(ABC):
    obj_name: str = None
    coord_types: Optional[tuple] = None

    def __init__(
        self,
        namespace: str,
        base_name: str,
        x: _Coord,
        y: _Coord,
        z: _Coord,
        nbt: NamedTag,
    ):
        assert isinstance(self.obj_name, str)
        assert self.coord_types is not None
        assert isinstance(namespace, str), "namespace must be a string"
        assert isinstance(base_name, str), "base_name must be a string"
        self._namespace = namespace
        self._base_name = base_name
        self._namespaced_name = None
        self._gen_namespaced_name()
        assert all(
            isinstance(c, self.coord_types) for c in (x, y, z)
        ), f"coordinates type must be in {self.coord_types}"
        self._x = self.coord_types[0](x)
        self._y = self.coord_types[0](y)
        self._z = self.coord_types[0](z)
        if isinstance(nbt, CompoundTag):
            self._nbt = NamedTag(nbt)
        elif isinstance(nbt, NamedTag):
            self._nbt = nbt
        else:
            raise Exception(f"nbt must be an NamedTag. Got {nbt}")

    def _gen_namespaced_name(self):
        self._namespaced_name = f'{self.namespace or ""}:{self.base_name}'

    def __repr__(self):
        return f"{self.obj_name}[{self.namespaced_name}, {self.x}, {self.y}, {self.z}]{{{self.nbt}}}"

    @property
    def namespaced_name(self) -> str:
        """
        The namespace:base_name of the entity represented by the object (eg: `minecraft:creeper`)

        If the given namespace is an empty string it will just return the base name.

        :return: The namespace:base_name of the block entity or just base_name if no namespace
        """
        return self._namespaced_name

    @namespaced_name.setter
    def namespaced_name(self, value: str):
        self._namespaced_name = value
        if ":" in value:
            self._namespace, self._base_name = value.split(":", 1)
        else:
            self._namespace, self._base_name = "", value

    @property
    def namespace(self) -> str:
        """
        The namespace of the block entity represented by the BlockEntity object (eg: `minecraft`)

        :return: The namespace of the block entity
        """
        return self._namespace

    @namespace.setter
    def namespace(self, value: str):
        self._namespace = value
        self._gen_namespaced_name()

    @property
    def base_name(self) -> str:
        """
        The base name of the block entity represented by the BlockEntity object (eg: `creeper`, `pig`)

        :return: The base name of the block entity
        """
        return self._base_name

    @base_name.setter
    def base_name(self, value: str):
        self._base_name = value
        self._gen_namespaced_name()

    @property
    def x(self) -> _Coord:
        return self._x

    @property
    def y(self) -> _Coord:
        return self._y

    @property
    def z(self) -> _Coord:
        return self._z

    @property
    def location(self) -> Union[BlockCoordinates, PointCoordinates]:
        return self._x, self._y, self._z

    @property
    def nbt(self) -> NamedTag:
        """
        The NBT behind the object

        :getter: Get the NBT data stored in the object
        :setter: Set the NBT data stored in the object

        :return: A NamedTag
        """
        return self._nbt

    @nbt.setter
    def nbt(self, value: NamedTag):
        if not isinstance(value, NamedTag):
            raise TypeError
        self._nbt = value



from typing import List
import numpy

from amulet_nbt import CompoundTag

from amulet.api.selection import SelectionBox
from amulet.api.data_types import BlockCoordinates, AnyNDArray


class MCStructureChunk:
    __slots__ = (
        "selection",
        "blocks",
        "palette",
        "shape",
        "entities",
        "block_entities",
    )

    def __init__(
        self,
        selection: SelectionBox,
        blocks: numpy.ndarray,
        palette: AnyNDArray,
        block_entities: List[CompoundTag],
        entities: List[CompoundTag],
    ):
        assert isinstance(blocks, numpy.ndarray)
        assert (
            selection.shape == blocks.shape
        ), "blocks shape does not match the specified section shape"
        self.selection = selection
        self.blocks = blocks
        self.palette = palette
        self.block_entities = block_entities
        self.entities = entities

    def __eq__(self, other):
        return (
            isinstance(other, MCStructureChunk)
            and self.selection == other.selection
            and self.shape == other.shape
            and numpy.array_equal(self.blocks, other.blocks)
            and self.entities == other.entities
            and self.block_entities == other.block_entities
        )

    @property
    def location(self) -> BlockCoordinates:
        return self.selection.min



from __future__ import annotations

from amulet.api.wrapper import EntityIDType
from .leveldb_8 import (
    LevelDB8Interface,
)


class LevelDB9Interface(LevelDB8Interface):

    chunk_version = 9

    def __init__(self):
        super().__init__()
        # EntityIDType.int_id is present until at least v7. Not sure which was present for v8
        self._set_feature("entity_format", EntityIDType.namespace_str_identifier)


export = LevelDB9Interface



import sys
from typing import TYPE_CHECKING, Tuple, Generator, Optional
import numpy

from amulet.api.data_types import Dimension, BlockCoordinates, FloatTriplet
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.block import Block, UniversalAirBlock
from amulet.api.errors import ChunkDoesNotExist, ChunkLoadError
from amulet.api.chunk import Chunk
from amulet.api.registry import BlockManager
from amulet.utils.matrix import transform_matrix, displacement_matrix
import amulet.api.level

if TYPE_CHECKING:
    from .base_level import BaseLevel


def is_sub_block(skip_blocks: Tuple[Block, ...], b: Block) -> bool:
    """Is the Block `b` a sub-block of any block in skip_blocks."""
    for skip_block in skip_blocks:
        if skip_block.namespaced_name == b.namespaced_name:
            other_properties = b.properties
            if skip_block.properties == {
                key: other_properties[key] for key in skip_block.properties.keys()
            }:
                return True
    return False


def gen_paste_blocks(
    block_palette: BlockManager, skip_blocks: Tuple[Block, ...]
) -> numpy.ndarray:
    """Create a numpy bool array of all the blocks which should be pasted.

    :param block_palette: The block palette of the chunk.
    :param skip_blocks: Blocks to not copy. If a property is not defined it will match any value.
    :return: Bool array of which blocks to copy.
    """
    return numpy.invert(
        numpy.vectorize(lambda b: is_sub_block(skip_blocks, b))(block_palette.blocks)
    )


def clone(
    src_structure: "BaseLevel",
    src_dimension: Dimension,
    src_selection: SelectionGroup,
    dst_structure: "BaseLevel",
    dst_dimension: Dimension,
    dst_selection_bounds: SelectionGroup,
    location: BlockCoordinates,
    scale: FloatTriplet = (1.0, 1.0, 1.0),
    rotation: FloatTriplet = (0.0, 0.0, 0.0),
    include_blocks: bool = True,
    include_entities: bool = True,
    skip_blocks: Tuple[Block, ...] = (),
    copy_chunk_not_exist: bool = False,
) -> Generator[float, None, None]:
    """Clone the source object data into the destination object with an optional transform.
    The src and dst can be the same object.
    Note this command may change in the future. Refer to all keyword arguments via the keyword.
    :param src_structure: The source structure to paste into the destination structure.
    :param src_dimension: The dimension of the source structure to use.
    :param src_selection: The area of the source structure to copy.
    :param dst_structure: The destination structure to paste into.
    :param dst_dimension: The dimension of the destination structure to use.
    :param dst_selection_bounds: The area of the destination structure that can be modified.
    :param location: The location where the centre of the `src_structure` will be in the `dst_structure`
    :param scale: The scale in the x, y and z axis. These can be negative to mirror.
    :param rotation: The rotation in degrees around each of the axis.
    :param include_blocks: Include blocks from the `src_structure`.
    :param include_entities: Include entities from the `src_structure`.
    :param skip_blocks: If a block matches a block in this list it will not be copied.
    :param copy_chunk_not_exist: If a chunk does not exist in the source should it be copied over as air. Always False where `src_structure` is a World.
    :return: A generator of floats from 0 to 1 with the progress of the paste operation.
    """
    location = tuple(location)
    src_selection = src_selection.merge_boxes()
    if include_blocks or include_entities:
        # we actually have to do something
        if isinstance(src_structure, amulet.api.level.World):
            copy_chunk_not_exist = False

        # TODO: look into if this can be a float so it will always be the exact middle
        rotation_point: numpy.ndarray = (
            (src_selection.max_array + src_selection.min_array) // 2
        ).astype(int)

        if src_structure is dst_structure and src_dimension == dst_dimension:
            # copying from an object to itself in the same dimension.
            # if the selections do not overlap this can be achieved directly
            # if they do overlap the selection will first need extracting
            # TODO: implement the above
            if (
                tuple(rotation_point) == location
                and scale == (1.0, 1.0, 1.0)
                and rotation == (0.0, 0.0, 0.0)
            ):
                # The src_object was pasted into itself at the same location. Nothing will change so do nothing.
                return
            src_structure = src_structure.extract_structure(
                src_selection, src_dimension
            )
            src_dimension = src_structure.dimensions[0]

        src_structure: "BaseLevel"

        # TODO: I don't know if this is feasible for large boxes: get the intersection of the source and destination selections and iterate over that to minimise work
        if any(rotation) or any(s != 1 for s in scale):
            # if the selection needs transforming
            rotation_radians = tuple(numpy.radians(rotation))
            transform = numpy.matmul(
                transform_matrix(scale, rotation_radians, location),
                displacement_matrix(*-rotation_point),
            )

            last_src: Optional[Tuple[int, int]] = None
            src_chunk: Optional[
                Chunk
            ] = None  # None here means the chunk does not exist or failed to load. Treat it as if it was air.
            last_dst: Optional[Tuple[int, int]] = None
            dst_chunk: Optional[
                Chunk
            ] = None  # None here means the chunk failed to load. Do not modify it.

            sum_progress = 0
            volumes = tuple(
                box.sub_chunk_count() for box in src_selection.selection_boxes
            )
            sum_volumes = sum(volumes)
            volumes = tuple(vol / sum_volumes for vol in volumes)

            if include_blocks:
                blocks_to_skip = set(skip_blocks)
                for box_index, box in enumerate(src_selection.selection_boxes):
                    for progress, src_coords, dst_coords in box.transformed_points(
                        transform
                    ):
                        if src_coords is not None:
                            dst_cx, dst_cy, dst_cz = dst_coords[0] >> 4
                            if (dst_cx, dst_cz) != last_dst:
                                last_dst = dst_cx, dst_cz
                                try:
                                    dst_chunk = dst_structure.get_chunk(
                                        dst_cx, dst_cz, dst_dimension
                                    )
                                except ChunkDoesNotExist:
                                    dst_chunk = dst_structure.create_chunk(
                                        dst_cx, dst_cz, dst_dimension
                                    )
                                except ChunkLoadError:
                                    dst_chunk = None

                            src_coords = numpy.floor(src_coords).astype(int)
                            # due to how the coords are found dst_coords will all be in the same sub-chunk
                            src_chunk_coords = src_coords >> 4

                            # split the src coords into which sub-chunks they came from
                            unique_chunks, inverse, counts = numpy.unique(
                                src_chunk_coords,
                                return_inverse=True,
                                return_counts=True,
                                axis=0,
                            )
                            chunk_indexes = numpy.argsort(inverse)
                            src_block_locations = numpy.split(
                                src_coords[chunk_indexes], numpy.cumsum(counts)[:-1]
                            )
                            dst_block_locations = numpy.split(
                                dst_coords[chunk_indexes], numpy.cumsum(counts)[:-1]
                            )
                            for chunk_location, src_blocks, dst_blocks in zip(
                                unique_chunks, src_block_locations, dst_block_locations
                            ):
                                # for each src sub-chunk
                                src_cx, src_cy, src_cz = chunk_location
                                if (src_cx, src_cz) != last_src:
                                    last_src = src_cx, src_cz
                                    try:
                                        src_chunk = src_structure.get_chunk(
                                            src_cx, src_cz, src_dimension
                                        )
                                    except ChunkLoadError:
                                        src_chunk = None

                                if dst_chunk is not None:
                                    if (
                                        src_chunk is not None
                                        and src_cy in src_chunk.blocks
                                    ):
                                        # TODO implement support for individual block rotation
                                        block_ids = src_chunk.blocks.get_sub_chunk(
                                            src_cy
                                        )[tuple(src_blocks.T % 16)]

                                        for block_id in numpy.unique(block_ids):
                                            block = src_chunk.block_palette[block_id]
                                            if not is_sub_block(skip_blocks, block):
                                                mask = block_ids == block_id
                                                dst_blocks_ = dst_blocks[mask]

                                                transformed_block = src_structure.translation_manager.transform_universal_block(
                                                    block, transform
                                                )

                                                dst_chunk.blocks.get_sub_chunk(dst_cy)[
                                                    tuple(dst_blocks_.T % 16)
                                                ] = dst_chunk.block_palette.get_add_block(
                                                    transformed_block
                                                )

                                                src_blocks_ = src_blocks[mask]
                                                for src_location, dst_location in zip(
                                                    src_blocks_, dst_blocks_
                                                ):
                                                    src_location = tuple(
                                                        src_location.tolist()
                                                    )
                                                    dst_location = tuple(
                                                        dst_location.tolist()
                                                    )

                                                    if (
                                                        src_location
                                                        in src_chunk.block_entities
                                                    ):
                                                        dst_chunk.block_entities[
                                                            dst_location
                                                        ] = src_chunk.block_entities[
                                                            src_location
                                                        ].new_at_location(
                                                            *dst_location
                                                        )
                                                    elif (
                                                        dst_location
                                                        in dst_chunk.block_entities
                                                    ):
                                                        del dst_chunk.block_entities[
                                                            dst_location
                                                        ]

                                                dst_chunk.changed = True
                                    elif UniversalAirBlock not in blocks_to_skip:
                                        dst_chunk.blocks.get_sub_chunk(dst_cy)[
                                            tuple(dst_blocks.T % 16)
                                        ] = dst_chunk.block_palette.get_add_block(
                                            UniversalAirBlock
                                        )
                                        for location in dst_blocks:
                                            location = tuple(location.tolist())
                                            if location in dst_chunk.block_entities:
                                                del dst_chunk.block_entities[location]
                                        dst_chunk.changed = True
                        yield sum_progress + volumes[box_index] * progress
                    sum_progress += volumes[box_index]

        else:
            # the selection can be cloned as is
            # the transform from the structure location to the world location
            offset = numpy.asarray(location).astype(int) - rotation_point
            moved_min_location = src_selection.min_array + offset

            iter_count = len(
                list(
                    src_structure.get_moved_coord_slice_box(
                        src_dimension,
                        moved_min_location,
                        src_selection,
                        dst_structure.sub_chunk_size,
                        yield_missing_chunks=copy_chunk_not_exist,
                    )
                )
            )

            count = 0

            for (
                src_chunk,
                src_slices,
                src_box,
                (dst_cx, dst_cz),
                dst_slices,
                dst_box,
            ) in src_structure.get_moved_chunk_slice_box(
                src_dimension,
                moved_min_location,
                src_selection,
                dst_structure.sub_chunk_size,
                create_missing_chunks=copy_chunk_not_exist,
            ):
                src_chunk: Chunk
                src_slices: Tuple[slice, slice, slice]
                src_box: SelectionBox
                dst_cx: int
                dst_cz: int
                dst_slices: Tuple[slice, slice, slice]
                dst_box: SelectionBox

                # load the destination chunk
                try:
                    dst_chunk = dst_structure.get_chunk(dst_cx, dst_cz, dst_dimension)
                except ChunkDoesNotExist:
                    dst_chunk = dst_structure.create_chunk(
                        dst_cx, dst_cz, dst_dimension
                    )
                except ChunkLoadError:
                    count += 1
                    continue

                if include_blocks:
                    # a boolean array specifying if each index should be pasted.
                    paste_blocks = gen_paste_blocks(
                        src_chunk.block_palette, skip_blocks
                    )

                    # create a look up table converting the source block ids to the destination block ids
                    gab = numpy.vectorize(
                        dst_chunk.block_palette.get_add_block, otypes=[numpy.uint32]
                    )
                    lut = gab(src_chunk.block_palette.blocks)

                    # iterate through all block entities in the chunk and work out if the block is going to be overwritten
                    remove_block_entities = []
                    for block_entity_location in dst_chunk.block_entities.keys():
                        if block_entity_location in dst_box:
                            chunk_block_entity_location = (
                                numpy.array(block_entity_location) - offset
                            )
                            chunk_block_entity_location[[0, 2]] %= 16
                            if paste_blocks[
                                src_chunk.blocks[tuple(chunk_block_entity_location)]
                            ]:
                                remove_block_entities.append(block_entity_location)
                    for block_entity_location in remove_block_entities:
                        del dst_chunk.block_entities[block_entity_location]

                    # copy over the source block entities if the source block is supposed to be pasted
                    for (
                        block_entity_location,
                        block_entity,
                    ) in src_chunk.block_entities.items():
                        if block_entity_location in src_box:
                            chunk_block_entity_location = numpy.array(
                                block_entity_location
                            )
                            chunk_block_entity_location[[0, 2]] %= 16
                            if paste_blocks[
                                src_chunk.blocks[tuple(chunk_block_entity_location)]
                            ]:
                                dst_chunk.block_entities.insert(
                                    block_entity.new_at_location(
                                        *offset + block_entity_location
                                    )
                                )

                    try:
                        block_mask = src_chunk.blocks[src_slices]
                        mask = paste_blocks[block_mask]
                        dst_chunk.blocks[dst_slices][mask] = lut[
                            src_chunk.blocks[src_slices]
                        ][mask]
                        dst_chunk.changed = True
                    except IndexError as e:
                        locals_copy = locals().copy()
                        import traceback

                        numpy_threshold = numpy.get_printoptions()["threshold"]
                        numpy.set_printoptions(threshold=sys.maxsize)
                        with open("clone_error.log", "w") as f:
                            for k, v in locals_copy.items():
                                f.write(f"{k}: {v}\n\n")
                        numpy.set_printoptions(threshold=numpy_threshold)
                        raise IndexError(
                            f"Error pasting.\nPlease notify the developers and include the clone_error.log file.\n{e}"
                        ) from e

                if include_entities:
                    # TODO: implement pasting entities when we support entities
                    pass

                count += 1
                yield count / iter_count

        yield 1.0



from abc import ABC, abstractmethod
from typing import Union, Tuple, Generator, Iterable, Any

BaseType = Any


class BaseRegistry(ABC):
    @abstractmethod
    def __len__(self):
        raise NotImplementedError

    @abstractmethod
    def __contains__(self, item: Union[int, BaseType]):
        raise NotImplementedError

    @abstractmethod
    def __iter__(self) -> Iterable[BaseType]:
        raise NotImplementedError

    @abstractmethod
    def values(self) -> Tuple[BaseType]:
        raise NotImplementedError

    @abstractmethod
    def items(self) -> Generator[Tuple[int, BaseType], None, None]:
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, item):
        raise NotImplementedError

    @abstractmethod
    def register(self, item: BaseType) -> int:
        raise NotImplementedError



from __future__ import annotations

from typing import Dict
import logging

import numpy

from amulet_nbt import CompoundTag, LongArrayTag
from amulet.api.chunk import Chunk
from amulet.utils.world_utils import decode_long_array, encode_long_array

from .base_anvil_interface import ChunkPathType, ChunkDataType
from .anvil_1444 import Anvil1444Interface as ParentInterface

log = logging.getLogger(__name__)


class Anvil1466Interface(ParentInterface):
    """
    Added multiple height maps. Now stored in a compound.
    """

    HeightMap = None
    Heightmaps: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("Heightmaps", CompoundTag)],
        CompoundTag,
    )

    def __init__(self):
        super().__init__()
        self._set_feature("height_map", "C|V1")

    @staticmethod
    def minor_is_valid(key: int):
        return 1466 <= key < 1467

    def _decode_height(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        heights = self.get_layer_obj(data, self.Heightmaps, pop_last=True)
        chunk.misc["height_mapC"] = h = {}
        for key, value in heights.items():
            if isinstance(value, LongArrayTag):
                try:
                    h[key] = decode_long_array(
                        value.np_array,
                        256,
                        (height_cy << 4).bit_length(),
                        dense=self.LongArrayDense,
                    ).reshape((16, 16)) + (floor_cy << 4)
                except Exception as e:
                    log.warning(e)

    def _encode_height(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        maps = [
            "WORLD_SURFACE_WG",
            "OCEAN_FLOOR_WG",
            "MOTION_BLOCKING",
            "MOTION_BLOCKING_NO_LEAVES",
            "OCEAN_FLOOR",
        ]
        if self._features["height_map"] == "C|V1":  # 1466
            maps = ("LIQUID", "SOLID", "LIGHT", "RAIN")
        elif self._features["height_map"] == "C|V2":  # 1484
            maps.append("LIGHT_BLOCKING")
        elif self._features["height_map"] == "C|V3":  # 1503
            maps.append("LIGHT_BLOCKING")
            maps.append("WORLD_SURFACE")
        elif self._features["height_map"] == "C|V4":  # 1908
            maps.append("WORLD_SURFACE")
        else:
            raise Exception
        heightmaps_temp: Dict[str, numpy.ndarray] = chunk.misc.get("height_mapC", {})
        heightmaps = CompoundTag()
        for heightmap in maps:
            if (
                heightmap in heightmaps_temp
                and isinstance(heightmaps_temp[heightmap], numpy.ndarray)
                and heightmaps_temp[heightmap].size == 256
            ):
                heightmaps[heightmap] = LongArrayTag(
                    encode_long_array(
                        heightmaps_temp[heightmap].ravel() - (floor_cy << 4),
                        (height_cy << 4).bit_length(),
                        self.LongArrayDense,
                    )
                )
        self.set_layer_obj(data, self.Heightmaps, heightmaps)


export = Anvil1466Interface



from typing import Dict, Union, Iterable, Tuple, List, Optional
from amulet_nbt import NamedTag


class ChunkData(Dict[bytes, Optional[bytes]]):
    def __init__(
        self,
        chunk_data: Union[Dict[bytes, bytes], Iterable[Tuple[bytes, bytes]]] = (),
        *,
        entity_actor: Iterable[NamedTag] = (),
        unknown_actor: Iterable[NamedTag] = (),
    ):
        super().__init__(chunk_data)
        self._entity_actor = list(entity_actor)
        self._unknown_actor = list(unknown_actor)

    @property
    def entity_actor(self) -> List[NamedTag]:
        """
        A list of entity actor data.
        UniqueID is stripped out. internalComponents is stripped out if there is no other data.
        """
        return self._entity_actor

    @property
    def unknown_actor(self) -> List[NamedTag]:
        """
        A list of actor data that does not fit in the known actor types.
        UniqueID is stripped out.
        internalComponents.{StorageKeyComponent}.StorageKey is replaced with a blank string.
        All keys matching this pattern will get replaced with the real storage key when saving (at least one must exist)
        """
        return self._unknown_actor



from __future__ import annotations

from .anvil_1908 import Anvil1908Interface as ParentInterface
from amulet.api.chunk import StatusFormats


class Anvil1912Interface(ParentInterface):
    """
    Changed status enum values
    """

    def __init__(self):
        super().__init__()
        self._set_feature("status", StatusFormats.Java_14)

    @staticmethod
    def minor_is_valid(key: int):
        return 1912 <= key < 1934


export = Anvil1912Interface
